import numpy as np
import healpy as hp
from .configurations import Configs
from .routines import _get_local_cov, _EB_to_QU, _E_to_QU, _B_to_QU, obj_to_array, array_to_obj, _log, _get_bandwidths
from .saving import _save_compsep_products, _get_full_path_out
from .needlets import _get_nside_lmax_from_b_ell, _get_needlet_windows_, _needlet_filtering, _get_good_channels_nl
from .pilcs import get_pilc_cov, get_prilc_cov
from .gilcs import _standardize_gnilc_run, Cn_C_Cn, _get_gilc_m, get_nuisance_idx
from .seds import _get_CMB_SED
import scipy
from numpy import linalg as lg
from types import SimpleNamespace
import os
from typing import Any, Dict, Optional, Union, List
import sys

def gpilc(config: Configs, input_alms: SimpleNamespace, compsep_run: Dict[str, Any], **kwargs) -> Optional[SimpleNamespace]:
    """
    Perform Generalized Polarization Internal Linear Combination (GPILC) on input spherical harmonics 
    to reconstruct Galactic emission at the observed microwave frequencies.

    Parameters
    ----------
        config : Configs
            Configuration object containing settings for the GPILC run. It should include:
                - lmax : int, maximum multipole for the component separation.
                - nside : int, HEALPix resolution parameter for compsep products.
                - fwhm_out : float, full width at half maximum of the output beam in arcminutes.
                - pixel_window_out : bool, whether to apply a pixel window to the output maps.
                - field_out : str, desired output fields (e.g., "E", "B", "QU", "EB", "QU_E", "QU_B").
                - save_compsep_products : bool, whether to save component separation products.
                - return_compsep_products : bool, whether to return component separation products.
                - path_outputs : str, path to save the output files.
                - leakage_correction : str, type of leakage correction to apply (e.g., "recycling", "purify").

        input_alms: SimpleNamespace
            Input multifrequency alms associated to polarization. 
            Each attribute should be a numpy array of shape (n_channels, 2, n_alms, n_components),
            if both E- and B-modes are provided, (n_channels, n_alms, n_components) otherwise.
            It should contain:
            - total : np.ndarray, alms of the total input maps.
            - nuisance : Optional[np.ndarray], alms of the nuisance maps (if available).
            Alternatively to nuisance, it can contain:
            - cmb : np.ndarray, alms of the CMB maps.
            - noise : np.ndarray, alms of the noise maps.

        compsep_run: Dict[str, Any]
            Dictionary with component separation parameters. It should include:
            - domain : str, either "pixel" or "needlet" for the component separation domain.
            - channels_out : list, indices of the frequency channels to reconstruct with GILC. Default is all channels.
            - depro_cmb : Optional[Union[float, list, np.ndarray]], deprojection factor for CMB (scalar or per needlet band). 
                    Default is None.
            - m_bias : Optional[Union[float, list, np.ndarray]], if not zero, it will include m_bias more (if m_bias > 0) 
                    or less (if m_bias < 0) modes in the reconstructed GPILC maps. Default is 0.
                    It can be a list if different values are needed for different needlet bands.
            - cmb_nuisance : bool, whether to include CMB alms in the nuisance covariance. Default is True.
            - needlet_config: Dictionary containing needlet settings. Needed if domain is "needlet". It should include:
                - "needlet_windows": Type of needlet windows ('cosine', 'standard', 'mexican').
                - "ell_peaks": List of integers defining multipoles of peaks for needlet bands (required for 'cosine').
                - "width": Width of the needlet windows (required for 'standard' and 'mexican').
                - "merging_needlets": Integer or list of integers defining ranges of needlets to be merged.
            - mask : Optional[np.ndarray], mask to apply to the maps (if available).
            - cov_noise_debias : Optional[Union[float, list, np.ndarray]], noise covariance debiasing factor.
                    If set to a non-zero value, it will subtract a 'noise_debias' fraction of noise covariance 
                    from the input and nuisance covariance matrices.
            
        **kwargs: 
            Dictionary of additional keyword arguments to pass to healpy function 'map2alm'.


    Returns
    -------
        Optional[SimpleNamespace] 
            Output object with component-separated maps, if config.return_compsep_products is True.
    """
    valid_fields = ["E", "B", "QU", "EB", "QU_E", "QU_B"]

    if config.field_out not in valid_fields:
        raise ValueError(f"Invalid field_out for GPILC. Must be one of: {', '.join(valid_fields)}.")

    compsep_run = _standardize_gnilc_run(compsep_run, input_alms.total.shape[0], config.lmax)

    compsep_run["nuis_idx"] = get_nuisance_idx(input_alms, compsep_run, config.verbose)
    if np.any(np.array(compsep_run["cov_noise_debias"]) != 0.):
        if not hasattr(input_alms, "noise"):
            raise ValueError("The input_alms object must have 'noise'' attribute for debiasing the covariance.")
        compsep_run["noise_idx"] = 2 if hasattr(input_alms, "fgds") else 1

    output_maps = _gpilc(config, obj_to_array(input_alms), compsep_run, **kwargs)
    
    outputs = array_to_obj(output_maps, input_alms)

    del output_maps
    compsep_run.pop("nuis_idx", None)
    compsep_run.pop("noise_idx", None)

    if config.save_compsep_products:
        _save_compsep_products(config, outputs, compsep_run, nsim=compsep_run["nsim"])
    
    if config.return_compsep_products:
        return outputs
    return None

def fgd_P_diagnostic(config: Configs, input_alms: SimpleNamespace, compsep_run: Dict[str, Any], **kwargs) -> Optional[SimpleNamespace]:
    """
    Perform foreground diagnostic for polarization (P) on input spherical harmonic coefficients (alms).

    Parameters
    ----------
        config: Configs
            Configuration object with general settings. It includes:
                - lmax : int, maximum multipole for the component separation.
                - nside : int, HEALPix resolution parameter for compsep products.
                - fwhm_out : float, full width at half maximum of the output beam in arcminutes.
                - field_out : str, desired fields to be considered (e.g., "T", "E", "B", "TEB").
                - save_compsep_products : bool, whether to save diagnostic maps.
                - return_compsep_products : bool, whether to return diagnostic maps.
                - path_outputs : str, path to save the output files.
        
        input_alms: SimpleNamespace
            Input multifrequency alms of polarization fields.
            Each attribute should be a numpy array of shape (n_channels, 2, n_alms, n_components),
            if both E- and B-modes are provided, (n_channels, n_alms, n_components) otherwise.
            It should contain:
            - total : np.ndarray, alms of the total input maps.
            - nuisance : Optional[np.ndarray], alms of the nuisance maps (if available).
            Alternatively to nuisance, it can contain:
            - cmb : np.ndarray, alms of the CMB maps.
            - noise : np.ndarray, alms of the noise maps.
    
        compsep_run: Dict[str, Any]
            Dictionary with diagnostic parameters. It should include:
                - domain : str, either "pixel" or "needlet" for the component separation domain.
                - cmb_nuisance : bool, whether to include CMB alms in the nuisance covariance. Default is True.
                - needlet_config: Dictionary containing needlet settings. Needed if domain is "needlet". It should include:
                    - "needlet_windows": Type of needlet windows ('cosine', 'standard', 'mexican').
                    - "ell_peaks": List of integers defining multipoles of peaks for needlet bands (required for 'cosine').
                    - "width": Width of the needlet windows (required for 'standard' and 'mexican').
                    - "merging_needlets": Integer or list of integers defining ranges of needlets to be merged.
                - adapt_nside : bool, whether to adapt the nside based on the needlet windows. Default is False.
                - mask : Optional[np.ndarray], mask to apply to the maps (if available).
                - cov_noise_debias : Optional[Union[float, list, np.ndarray]], noise covariance debiasing factor.
        
        **kwargs: 
            Dictionary of additional keyword arguments to pass to healpy function 'map2alm'.

    Returns
    -------
        Optional[SimpleNamespace]
            Object containing foreground diagnostic maps.
    """

    valid_fields = ["E", "B", "QU", "EB", "QU_E", "QU_B"]
    if config.field_out not in valid_fields:
        raise ValueError(f"Invalid field_in for foreground diagnostic. Must be one of: {', '.join(valid_fields)}")

    compsep_run.setdefault("cmb_nuisance", True)

    compsep_run["nuis_idx"] = get_nuisance_idx(input_alms, compsep_run, config.verbose)
    if np.any(np.array(compsep_run["cov_noise_debias"]) != 0.):
        if not hasattr(input_alms, "noise"):
            raise ValueError("The input_alms object must have 'noise'' attribute for debiasing the covariance.")
        compsep_run["noise_idx"] = 2 if hasattr(input_alms, "fgds") else 1

    if isinstance(compsep_run["nuis_idx"], int):
        nuis_alms = (obj_to_array(input_alms))[...,compsep_run["nuis_idx"]]
    elif isinstance(compsep_run["nuis_idx"], list):
        nuis_alms = (obj_to_array(input_alms))[...,compsep_run["nuis_idx"][0]] + (obj_to_array(input_alms))[...,compsep_run["nuis_idx"][1]]
    inputs_alms_for_diagn = np.concatenate([
        input_alms.total[...,np.newaxis],
        nuis_alms[...,np.newaxis]],axis=-1)
    del nuis_alms

    if np.any(np.array(compsep_run["cov_noise_debias"]) != 0.):
        noi_alms = (obj_to_array(input_alms))[...,compsep_run["noise_idx"]]
        inputs_alms_for_diagn = np.concatenate([inputs_alms_for_diagn, noi_alms[...,np.newaxis]], axis=-1)
        del noi_alms

    output_maps = _fgd_P_diagnostic(config, inputs_alms_for_diagn, compsep_run)
    del inputs_alms_for_diagn
    
    outputs = SimpleNamespace()
    setattr(outputs, "m", output_maps)
    del output_maps

    compsep_run.pop("nuis_idx", None)
    compsep_run.pop("noise_idx", None)

    if config.save_compsep_products:
        _save_compsep_products(config, outputs, compsep_run, nsim=compsep_run["nsim"])
    
    if config.return_compsep_products:
        return outputs
    return None

def _gpilc(config: Configs, input_alms: np.ndarray, compsep_run: Dict[str, Any], **kwargs) -> np.ndarray:
    """
    Perform Generalized Polarization Internal Linear Combination (GPILC) on input spherical harmonics.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the GPILC run.
        input_alms: np.ndarray
            Input multifrequency alms associated to polarization. 
            Shape should be (n_channels, 2, n_alms, n_components) if both E- and B-modes are provided, or (n_channels, n_alms, n_components) otherwise.
        compsep_run: Dict[str, Any]
            Dictionary with component separation parameters. See `gpilc` function for details.
        **kwargs:
            Additional keyword arguments to pass to healpy function 'map2alm'.

    Returns
    -------
        np.ndarray
            Output maps after performing GPILC. Shape will depend on the input alms and compsep_run settings.
    """
    
    if compsep_run["domain"] == "pixel":
        return _gpilc_pixel(config, input_alms, compsep_run, **kwargs)
    elif compsep_run["domain"] == "needlet":
        return _gpilc_needlet(config, input_alms, compsep_run, **kwargs)
    else:
        raise ValueError(f"Invalid domain '{compsep_run['domain']}' for GPILC. Must be 'pixel' or 'needlet'.")

def _fgd_P_diagnostic(config: Configs, input_alms: np.ndarray, compsep_run: Dict[str, Any]) -> np.ndarray:
    """
    Perform foreground diagnostic for polarization (P) on input spherical harmonic coefficients (alms) based on domain configuration.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the diagnostic run. See `fgd_P_diagnostic` function for details.
        input_alms: np.ndarray
            Input multifrequency alms of polarization fields. 
            Shape should be (n_channels, 2, n_alms) if both E- and B-modes are provided, or (n_channels, n_alms) otherwise.
        compsep_run: Dict[str, Any]
            Dictionary with diagnostic parameters. See `fgd_P_diagnostic` function for details.
    
    Returns
    -------
        np.ndarray
            Foreground diagnostic maps. Shape will depend on the input alms and compsep_run settings.
    """

    if compsep_run["domain"] == "pixel":
        return _fgd_P_diagnostic_pixel(config, input_alms, compsep_run)
    elif compsep_run["domain"] == "needlet":
        return _fgd_P_diagnostic_needlet(config, input_alms, compsep_run)
    else:
        raise ValueError(f"Invalid domain '{compsep_run['domain']}' for foreground diagnostic. Must be 'pixel' or 'needlet'.")

def _gpilc_pixel(config: Configs, input_alms: np.ndarray, compsep_run: Dict[str, Any], **kwargs):
    """
    Perform Generalized Polarization Internal Linear Combination (GPILC) on input spherical harmonics in pixel space.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the GPILC run. See `gpilc` function for details.
        input_alms: np.ndarray
            Input multifrequency alms associated to polarization. 
            Shape should be (n_channels, 2, n_alms, n_components) if both E- and B-modes are provided, or (n_channels, n_alms, n_components) otherwise.
        compsep_run: Dict[str, Any]
            Dictionary with component separation parameters. See `gpilc` function for details.
        **kwargs:
            Additional keyword arguments to pass to healpy function 'map2alm'.

    Returns
    -------
        np.ndarray
            Output maps after performing GPILC in pixel space. Shape will depend on the input alms and compsep_run settings.
    """

    compsep_run["good_channels"] = _get_good_channels_nl(config, np.ones(config.lmax+1))
#    compsep_run["good_channels"] = np.arange(input_alms.shape[0])

    input_maps = np.zeros((compsep_run["good_channels"].shape[0], 2, 12 * config.nside**2, input_alms.shape[-1]))

    def alm_to_polmap(E=None, B=None):
        T = np.zeros_like(E if E is not None else B)
        return hp.alm2map([T, E if E is not None else T, B if B is not None else T],
                          config.nside, lmax=config.lmax, pol=True)[1:]

    for n, channel in enumerate(compsep_run["good_channels"]):
        for c in range(input_alms.shape[-1]):
            if input_alms.ndim == 4:
                input_maps[n, ..., c] = alm_to_polmap(E=input_alms[channel, 0, :, c],
                                                      B=input_alms[channel, 1, :, c])
            elif input_alms.ndim == 3:
                if config.field_out in ["QU_E", "E"]:
                    input_maps[n, ..., c] = alm_to_polmap(E=input_alms[channel, :, c])
                elif config.field_out in ["QU_B", "B"]:
                    input_maps[n, ..., c] = alm_to_polmap(B=input_alms[channel, :, c])

    # Perform GPILC separation
    output_maps = _gpilc_maps(
        config, input_maps, compsep_run,
        np.ones(config.lmax + 1), depro_cmb=compsep_run["depro_cmb"],
        m_bias=compsep_run["m_bias"], noise_debias=compsep_run["cov_noise_debias"],
    )

    if (config.field_out in ["QU", "QU_E", "QU_B"] and config.pixel_window_out) or (config.field_out in ["E","B"]) or (config.field_out=="EB"):
        for f, c in np.ndindex(output_maps.shape[0],output_maps.shape[-1]):
#            if ("mask" in compsep_run) and (config.mask_type == "observed_patch"):
#                if "_purify" in config.leakage_correction:
#                    alm_out = purify_master(output_maps[f,...,c], compsep_run["mask"], config.lmax,
#                                            purify_E=("E" in config.leakage_correction))
#                    alm_out = np.concatenate([(0.*alm_out[0])[np.newaxis],alm_out], axis=0)
#                elif "_recycling" in config.leakage_correction:
#                    iterations = int(re.search(r'iterations(\d+)', config.leakage_correction).group(1)) \
#                        if "_iterations" in config.leakage_correction else 0
#                    alm_out = purify_recycling(output_maps[f, ..., c],
#                                               output_maps[f, ..., 0],
#                                               compsep_run["mask"],
#                                               config.lmax,
#                                               purify_E=("E" in config.leakage_correction),
#                                               iterations=iterations, **kwargs)
#                    alm_out = np.concatenate([(0.*alm_out[0])[np.newaxis],alm_out], axis=0)
#               else:
#                   alm_out = hp.map2alm(np.array([0. * output_maps[f, 0, :, c],
#                                            output_maps[f, 0, :, c],
#                                            output_maps[f, 1, :, c]]) * compsep_run["mask"],lmax=config.lmax, pol=True, **kwargs)
#            else:
            alm_out = hp.map2alm([0. * output_maps[f, 0, :, c],
                                        output_maps[f, 0, :, c],
                                        output_maps[f, 1, :, c]],
                                        lmax=config.lmax, pol=True, **kwargs)

            if (config.field_out in ["QU", "QU_E", "QU_B"]):
                output_maps[f,...,c] = hp.alm2map(alm_out, config.nside, lmax=config.lmax, pol=True,
                                                    pixwin=config.pixel_window_out)[1:]
            elif config.field_out in ["E","B"]:
                output_maps[f,0,:,c] = hp.alm2map(alm_out[1] if config.field_out=="E" else alm_out[2],
                                                    config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)
            elif config.field_out=="EB":
                output_maps[f,0,:,c] = hp.alm2map(alm_out[1], config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)
                output_maps[f,1,:,c] = hp.alm2map(alm_out[2], config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)

        if config.field_out in ["E","B"]:
            output_maps = output_maps[:,0]

    del compsep_run['good_channels']
    if "mask" in compsep_run:
        if output_maps.ndim == 3:
            output_maps[:,compsep_run["mask"] == 0.,:] = 0.
        elif output_maps.ndim == 4:
            output_maps[...,compsep_run["mask"] == 0.,:] = 0.

    return output_maps

def _fgd_P_diagnostic_pixel(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: Dict[str, Any]
) -> np.ndarray:
    """
    Perform foreground diagnostic for polarization (P) on input spherical harmonic coefficients (alms) in pixel space.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the diagnostic run. See `fgd_P_diagnostic` function for details.
        input_alms: np.ndarray
            Input multifrequency alms of polarization fields.
            Shape should be (n_channels, 2, n_alms) if both E- and B-modes are provided, or (n_channels, n_alms) otherwise.
        compsep_run: Dict[str, Any]
            Dictionary with diagnostic parameters. See `fgd_P_diagnostic` function for details.

    Returns
    -------
        np.ndarray
            Foreground diagnostic maps in pixel space. Shape will depend on the input alms and compsep_run settings.

    """
    compsep_run["good_channels"] = _get_good_channels_nl(config, np.ones(config.lmax+1))
    
    input_maps = np.zeros((compsep_run["good_channels"].shape[0], 2, 12 * config.nside**2, input_alms.shape[-1]))
    
    def alm_to_polmap(E=None, B=None):
        T = np.zeros_like(E if E is not None else B)
        return hp.alm2map(np.ascontiguousarray(np.array([T, E if E is not None else T, B if B is not None else T])),
                          config.nside, lmax=config.lmax, pol=True)[1:]

    for n, channel in enumerate(compsep_run["good_channels"]):
        for c in range(input_alms.shape[-1]):
            if input_alms.ndim == 4:
                input_maps[n, ..., c] = alm_to_polmap(E=input_alms[channel, 0, :, c],
                                                      B=input_alms[channel, 1, :, c])
            elif input_alms.ndim == 3:
                if config.field_out in ["QU_E", "E"]:
                    input_maps[n, ..., c] = alm_to_polmap(E=input_alms[channel, :, c])
                elif config.field_out in ["QU_B", "B"]:
                    input_maps[n, ..., c] = alm_to_polmap(B=input_alms[channel, :, c])

    output_maps = _fgd_P_diagnostic_maps(config, input_maps, compsep_run, np.ones(config.lmax+1), noise_debias=compsep_run["cov_noise_debias"])

    if "mask" in compsep_run:
        output_maps[compsep_run["mask"] == 0.] = 0.
        
    return output_maps

def _gpilc_needlet(config: Configs,
                   input_alms: np.ndarray,
                   compsep_run: Dict[str, Any],
                   **kwargs) -> np.ndarray:
    """
    Perform Generalized Polarization Internal Linear Combination (GPILC) on input spherical harmonics in needlet space.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the GPILC run. See `gpilc` function for details.
        input_alms: np.ndarray
            Input multifrequency alms associated to polarization.
            Shape should be (n_channels, 2, n_alms, n_components) if both E- and B-modes are provided, or (n_channels, n_alms, n_components) otherwise.
        compsep_run: Dict[str, Any]
            Dictionary with component separation parameters. See `gpilc` function for details.
        **kwargs:
            Additional keyword arguments to pass to healpy function 'map2alm'.
    
    Returns
    -------
        np.ndarray
            Output maps after performing GPILC in needlet space. Shape will depend on the input alms and compsep_run settings.
    """

    b_ell = _get_needlet_windows_(compsep_run["needlet_config"], config.lmax)
    b_ell = b_ell**2

    if compsep_run['save_needlets']:
        compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        os.makedirs(compsep_run["path_out"], exist_ok=True)
        np.save(os.path.join(compsep_run["path_out"], "needlet_bands"), b_ell)

    output_maps = np.zeros((len(compsep_run["channels_out"]), 2, hp.nside2npix(config.nside), input_alms.shape[-1]))
    
    for j, b_ell_j in enumerate(b_ell):
        output_maps += _gpilc_needlet_j(config, input_alms, compsep_run,
                                        b_ell_j,
                                        depro_cmb=compsep_run["depro_cmb"][j],
                                        m_bias=compsep_run["m_bias"][j],
                                        noise_debias=compsep_run["cov_noise_debias"][j],
                                        )

    
    if ((config.field_out in ["QU", "QU_E", "QU_B"]) and config.pixel_window_out) or (config.field_out in ["E","B"]) or (config.field_out=="EB"):
        for f, c in np.ndindex(output_maps.shape[0],output_maps.shape[-1]):
#            if ("mask" in compsep_run) and (config.mask_type == "observed_patch") and config.leakage_correction is not None:
#                if "_purify" in config.leakage_correction:
#                    alm_out = purify_master(output_maps[f, ..., c],
#                                            compsep_run["mask"],
#                                            config.lmax,
#                                            purify_E=("E" in config.leakage_correction))
#                    alm_out = np.concatenate([(0.*alm_out[0])[np.newaxis],alm_out], axis=0)
#                elif "_recycling" in config.leakage_correction:
#                    iterations = int(re.search(r'iterations(\d+)', config.leakage_correction).group(1)) \
#                        if "_iterations" in config.leakage_correction else 0
#                    alm_out = purify_recycling(output_maps[f, ..., c],
#                                            output_maps[f, ..., 0],
#                                            compsep_run["mask"],
#                                            config.lmax,
#                                            purify_E=("E" in config.leakage_correction),
#                                            iterations=iterations, **kwargs)
#                    alm_out = np.concatenate([(0.*alm_out[0])[np.newaxis],alm_out], axis=0)
#                elif config.leakage_correction=="mask_only":
#                    alm_out = hp.map2alm(np.array([0. * output_maps[f, 0, :, c],
#                                            output_maps[f, 0, :, c],
#                                            output_maps[f, 1, :, c]]) * compsep_run["mask"],lmax=config.lmax, pol=True, **kwargs)
#            else:
            alm_out = hp.map2alm([0. * output_maps[f, 0, :, c],
                                        output_maps[f, 0, :, c],
                                        output_maps[f, 1, :, c]],
                                        lmax=config.lmax, pol=True, **kwargs)

            if (config.field_out in ["QU", "QU_E", "QU_B"]):
                output_maps[f, ..., c] = hp.alm2map(alm_out,
                                                    config.nside,
                                                    lmax=config.lmax,
                                                    pol=True,
                                                    pixwin=config.pixel_window_out)[1:]
            elif config.field_out in ["E","B"]:
                output_maps[f,0,:,c] = hp.alm2map(alm_out[1] if config.field_out=="E" else alm_out[2], config.nside, lmax=config.lmax,
                                                     pol=False, pixwin=config.pixel_window_out)
            elif config.field_out=="EB":
                output_maps[f, 0, :, c] = hp.alm2map(alm_out[1], config.nside, lmax=config.lmax,
                                                     pol=False, pixwin=config.pixel_window_out)
                output_maps[f, 1, :, c] = hp.alm2map(alm_out[2], config.nside, lmax=config.lmax,
                                                     pol=False, pixwin=config.pixel_window_out)

        if config.field_out in ["E","B"]:
            output_maps = output_maps[:,0]

    if "mask" in compsep_run:
        if output_maps.ndim == 3:
            output_maps[:,compsep_run["mask"] == 0.,:] = 0.
        elif output_maps.ndim == 4:
            output_maps[...,compsep_run["mask"] == 0.,:] = 0.
    
    return output_maps
        
def _gpilc_needlet_j(config: Configs,
                     input_alms: np.ndarray,
                     compsep_run: Dict[str, Any],
                     b_ell: np.ndarray,
                     depro_cmb: Optional[float] = None,
                     m_bias: Optional[Union[int, float]] = 0,
                     noise_debias: Optional[float] = 0.,
                     ) -> np.ndarray:
    """
    Perform GPILC on a single needlet band.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the GPILC run. See `gpilc` function for details.
        input_alms: np.ndarray
            Input multifrequency alms associated to polarization.
            Shape should be (n_channels, 2, n_alms, n_components) if both E- and B-modes are provided, or (n_channels, n_alms, n_components) otherwise.
        compsep_run: Dict[str, Any]
            Dictionary with component separation parameters. See `gpilc` function for details.
        b_ell: np.ndarray
            Needlet bandpass filter. Shape should be (lmax+1).
        depro_cmb: float, optional
            Deprojection factor for CMB. If None, no deprojection is applied. 
            Otherwise CMB residuals in GPILC maps will be at the level of depro_cmb * CMB_input in the considered needlet band.
        m_bias: int or float, optional
            It will include m_bias more (if m_bias > 0) or less (if m_bias < 0) modes in the reconstructed GNILC maps.
        noise_debias: float, optional
            Noise debiasing factor. If set to a non-zero value, it will subtract a 'noise_debias' fraction of
            noise covariance from the input and nuisance covariance matrices.
    
    Returns
    -------
        np.ndarray
            Output maps after performing GPILC on the specified needlet band. Shape will depend on the input alms and compsep_run settings.
    """

    nside_, lmax_ = config.nside, config.lmax

    compsep_run["good_channels"] = _get_good_channels_nl(config, b_ell)
#    compsep_run["good_channels"] = np.arange(input_alms.shape[0])

    input_maps_nl = np.zeros((compsep_run["good_channels"].shape[0], 2, 12 * nside_**2, input_alms.shape[-1]))

    for n, channel in enumerate(compsep_run["good_channels"]):
        input_alms_j = np.zeros((2, hp.Alm.getsize(lmax_), input_alms.shape[-1]), dtype=complex)
        if input_alms.ndim == 4:
            for k in range(2):
                input_alms_j[k] = _needlet_filtering(input_alms[channel,k], b_ell, lmax_)
        elif input_alms.ndim == 3:
            if config.field_out in ["QU_E", "E"]:
                input_alms_j[0] = _needlet_filtering(input_alms[channel], b_ell, lmax_)
            elif config.field_out in ["QU_B", "B"]:
                input_alms_j[1] = _needlet_filtering(input_alms[channel], b_ell, lmax_)
        
#        if ("mask" in compsep_run) and (config.mask_type == "observed_patch"):
#            for c in range(input_alms.shape[-1]):
#                input_maps_nl[n,...,c] = (hp.alm2map(np.ascontiguousarray([0. * input_alms_j[0, :, c],
#                                                    input_alms_j[0, :, c],
#                                                    input_alms_j[1, :, c]]),
#                                                   nside_, lmax=lmax_, pol=True)[1:]) * compsep_run["mask"]       
#        else:
        for c in range(input_alms.shape[-1]):
            input_maps_nl[n, ..., c] = hp.alm2map(np.ascontiguousarray([0. * input_alms_j[0, :, c],
                                                input_alms_j[0, :, c],
                                                input_alms_j[1, :, c]]),
                                                nside_, lmax=lmax_, pol=True)[1:]

    output_maps_nl = _gpilc_maps(config, input_maps_nl, compsep_run,
                              b_ell, depro_cmb=depro_cmb, m_bias=m_bias, noise_debias=noise_debias)

    del input_maps_nl
    del compsep_run['good_channels']

    return output_maps_nl

def _fgd_P_diagnostic_needlet(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict
) -> np.ndarray:
    """
    Perform diagnostic of foreground complexity in polarization intensity P and in needlet space.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the diagnostic run. See `fgd_P_diagnostic` function for details.
        input_alms: np.ndarray
            Input multifrequency alms of polarization fields.
            Shape should be (n_channels, 2, n_alms) if both E- and B-modes are provided, or (n_channels, n_alms) otherwise.
        compsep_run: Dict[str, Any]
            Dictionary with diagnostic parameters. See `fgd_P_diagnostic` function for details.
    
    Returns
    -------
        np.ndarray
            Foreground diagnostic maps in needlet space. Shape will depend on the input alms and compsep_run settings.
    """

    b_ell = _get_needlet_windows_(compsep_run["needlet_config"], config.lmax)
    b_ell = b_ell**2

    if compsep_run['save_needlets']:
        compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        os.makedirs(compsep_run["path_out"], exist_ok=True)
        np.save(os.path.join(compsep_run["path_out"], "needlet_bands"), b_ell)

    output_maps = np.zeros((b_ell.shape[0], hp.nside2npix(config.nside)))
    
    for j in range(b_ell.shape[0]):
        output_maps[j] = _fgd_P_diagnostic_needlet_j(config, input_alms, compsep_run, b_ell[j], 
                                                      noise_debias=compsep_run["cov_noise_debias"][j])
    
    if "mask" in compsep_run:
        output_maps[:,compsep_run["mask"] == 0.] = 0.
        
    return output_maps
        
def _fgd_P_diagnostic_needlet_j(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    b_ell: np.ndarray,
    noise_debias: Optional[float] = 0.0
) -> np.ndarray:
    """
    Perform diagnostic of foreground complexity in polarization intensity P for a single needlet band.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the diagnostic run. See `fgd_P_diagnostic` function for details.
        input_alms: np.ndarray
            Input multifrequency alms of polarization fields.
            Shape should be (n_channels, 2, n_alms) if both E- and B-modes are provided, or (n_channels, n_alms) otherwise.
        compsep_run: Dict[str, Any]
            Dictionary with diagnostic parameters. See `fgd_P_diagnostic` function for details.
        b_ell: np.ndarray
            Needlet bandpass filter for the current band. Shape should be (lmax+1).
        noise_debias: float, optional
            Noise debiasing factor. If set to a non-zero value, it will subtract a 'noise_debias' fraction of
            noise covariance from the input and nuisance covariance matrices.
    
    Returns
    -------
        np.ndarray
            Foreground diagnostic maps in needlet space for the specified band. Shape will depend on the input alms and compsep_run settings.
    """

    if "mask" in compsep_run or not compsep_run["adapt_nside"]:
        nside_, lmax_ = config.nside, config.lmax
    else:
        nside_, lmax_ = _get_nside_lmax_from_b_ell(b_ell,config.nside,config.lmax)
        
    compsep_run["good_channels"] = _get_good_channels_nl(config, b_ell)
#    compsep_run["good_channels"] = np.arange(input_alms.shape[0])

    input_maps_nl = np.zeros((compsep_run["good_channels"].shape[0], 2, 12 * nside_**2, input_alms.shape[-1]))

    for n, channel in enumerate(compsep_run["good_channels"]):
        input_alms_j = np.zeros((2, hp.Alm.getsize(lmax_), input_alms.shape[-1]), dtype=complex)
        if input_alms.ndim == 4:
            for k in range(2):
                input_alms_j[k] = _needlet_filtering(input_alms[channel,k], b_ell, lmax_)
        elif input_alms.ndim == 3:
            if config.field_out in ["QU_E", "E"]:
                input_alms_j[0] = _needlet_filtering(input_alms[channel], b_ell, lmax_)
            elif config.field_out in ["QU_B", "B"]:
                input_alms_j[1] = _needlet_filtering(input_alms[channel], b_ell, lmax_)
        
        for c in range(input_alms.shape[-1]):
            input_maps_nl[n, ..., c] = hp.alm2map(np.ascontiguousarray([0. * input_alms_j[0, :, c],
                                                input_alms_j[0, :, c],
                                                input_alms_j[1, :, c]]),
                                                nside_, lmax=lmax_, pol=True)[1:]

    output_maps_nl = _fgd_P_diagnostic_maps(config, input_maps_nl, compsep_run, b_ell, noise_debias=noise_debias)
    del input_maps_nl
    del compsep_run['good_channels']

    if hp.get_nside(output_maps_nl) < config.nside:
        output_maps_nl = hp.ud_grade(output_maps_nl, config.nside)

    return output_maps_nl

def _gpilc_maps(
    config: Configs,
    input_maps: np.ndarray,
    compsep_run: dict,
    b_ell: np.ndarray,
    depro_cmb: Optional[float] = None,
    m_bias: Optional[Union[int, float]] = 0,
    noise_debias: Optional[float] = 0.0
) -> np.ndarray:
    """
    Perform Generalized Polarization Internal Linear Combination (GPILC) on input maps.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the GPILC run. See `gpilc` function for details.
        input_maps: np.ndarray
            Input multifrequency maps associated to polarization.
            Shape should be (n_channels, 2, n_pixels, n_components).
        compsep_run: dict
            Dictionary with component separation parameters. See `gpilc` function for details.
        b_ell: np.ndarray
            Bandpass filter for the GPILC. Shape should be (lmax+1).
            If compsep_run["domain"] is "pixel", it should be an array of ones.
        depro_cmb: float, optional
            Deprojection factor for CMB. If None, no deprojection is applied.
            Otherwise CMB residuals in GPILC maps will be at the level of depro_cmb * CMB_input.
        m_bias: int or float, optional
            It will include m_bias more (if m_bias > 0) or less (if m_bias < 0) modes in the reconstructed GPILC maps.
        noise_debias: float, optional
            Noise debiasing factor. If set to a non-zero value, it will subtract a 'noise_debias' fraction of
            noise covariance from the input and nuisance covariance matrices.
    
    Returns
    -------
        np.ndarray
            Output maps after performing GPILC. Shape is (n_channels_out, 2, n_pixels, n_components),
            where n_channels_out is the number of channels specified in compsep_run["channels_out"].
    """

    cov = (get_prilc_cov(input_maps[...,0], config.lmax, compsep_run, b_ell)).T

    if isinstance(compsep_run["nuis_idx"], int):
        cov_n = (get_prilc_cov(input_maps[...,compsep_run["nuis_idx"]], config.lmax, compsep_run, b_ell)).T
    elif isinstance(compsep_run["nuis_idx"], list):
        cov_n = (get_prilc_cov(input_maps[...,compsep_run["nuis_idx"][0]] + input_maps[...,compsep_run["nuis_idx"][1]], config.lmax, compsep_run, b_ell)).T

    if noise_debias != 0.:
        cov_noi = (get_prilc_cov(input_maps[...,compsep_run["noise_idx"]], config.lmax, compsep_run, b_ell)).T
        cov = cov - noise_debias * cov_noi
        cov_n = cov_n - noise_debias * cov_noi
        del cov_noi

    λ, U = Cn_C_Cn(cov,cov_n)
    λ[λ<1.]=1.

    W = _get_gpilc_weights(
        config, U, λ, cov, cov_n, input_maps.shape, compsep_run,
        depro_cmb=depro_cmb, m_bias=m_bias
    )

    del cov, cov_n, U, λ

    if compsep_run["ilc_bias"] == 0.:
        if W.ndim==2:
            output_maps = np.einsum('li,ifjk->lfjk', W, input_maps)
        elif W.ndim==3:
            output_maps = np.zeros((W.shape[0], 2, *input_maps.shape[-2:]))
            output_maps[:, 0] = (
                np.einsum('li,ijk->ljk', W[0], input_maps[:, 0]) -
                np.einsum('li,ijk->ljk', W[1], input_maps[:, 1])
            )
            output_maps[:, 1] = (
                np.einsum('li,ijk->ljk', W[1], input_maps[:, 0]) +
                np.einsum('li,ijk->ljk', W[0], input_maps[:, 1])
            )
    else:
        if W.ndim==3:
            output_maps = np.einsum('jli,ifjk->lfjk', W, input_maps)
        elif W.ndim==4:
            output_maps = np.zeros((W.shape[0], 2, *input_maps.shape[-2:]))
            output_maps[:, 0] = (
                np.einsum('jli,ijk->ljk', W[0], input_maps[:, 0]) -
                np.einsum('jli,ijk->ljk', W[1], input_maps[:, 1])
            )
            output_maps[:, 1] = (
                np.einsum('jli,ijk->ljk', W[1], input_maps[:, 0]) +
                np.einsum('jli,ijk->ljk', W[0], input_maps[:, 1])
            )

    outputs = []
    for channel in compsep_run["channels_out"]:
        if channel in compsep_run["good_channels"]:
            outputs.append(output_maps[compsep_run["good_channels"] == channel][0])
        else:
            outputs.append(np.zeros(output_maps.shape[1:]))
    del output_maps

    return np.array(outputs)

def _fgd_P_diagnostic_maps(
    config: Configs,
    input_maps: np.ndarray,
    compsep_run: dict,
    b_ell: np.ndarray,
    noise_debias: Optional[float] = 0.0
) -> np.ndarray:
    """
    Perform diagnostic of foreground complexity in polarization intensity P.

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the diagnostic run. See `fgd_P_diagnostic` function for details.
        input_maps: np.ndarray
            Input multifrequency maps associated to polarization.
            Shape should be (n_channels, 2, n_pixels).
        compsep_run: dict
            Dictionary with component separation parameters. See `fgd_P_diagnostic` function for details.
        b_ell: np.ndarray
            Needlet bandpass filter for the diagnostic. Shape should be (lmax+1).
            If compsep_run["domain"] is "pixel", it should be an array of ones.
        noise_debias: float, optional
            Noise debiasing factor. If set to a non-zero value, it will subtract a 'noise_debias' fraction of
            noise covariance from the input and nuisance covariance matrices.
    
    Returns
    -------
        np.ndarray
            Maps of foreground complexity in polarization intensity P.
            Shape is (n_pixels).
    """

    cov = (get_prilc_cov(input_maps[...,0], config.lmax, compsep_run, b_ell)).T
    cov_n = (get_prilc_cov(input_maps[...,1], config.lmax, compsep_run, b_ell)).T
    
    if noise_debias != 0.:
        cov_noi = (get_prilc_cov(input_maps[...,2], config.lmax, compsep_run, b_ell)).T
        cov = cov - noise_debias * cov_noi
        cov_n = cov_n - noise_debias * cov_noi
        del cov_noi

    λ, U = Cn_C_Cn(cov,cov_n)
    λ[λ<1.]=1.

    m = _get_gilc_m(λ)

    if isinstance(m, (int, float)):
        m = np.repeat(m, input_maps.shape[-2])

    return m

def _get_gpilc_weights(
    config: Configs,
    U: np.ndarray,
    λ: np.ndarray,
    cov: np.ndarray,
    cov_n: np.ndarray,
    input_shapes: tuple,
    compsep_run: dict,
    depro_cmb: Optional[float] = None,
    m_bias: Optional[Union[int, float]] = 0
) -> np.ndarray:
    """
    Compute the weights for Generalized Polarization Internal Linear Combination (GPILC).

    Parameters
    ----------
        config: Configs
            Configuration object with settings for the GPILC run. See `gpilc` function for details.
        U: np.ndarray
            Eigenvectors of the covariance matrix.
        λ: np.ndarray
            Eigenvalues of the covariance matrix.
        cov: np.ndarray
            Covariance matrix of the input maps.
            Shape should be (n_pixels, n_channels, n_channels) for 3D covariance or (n_channels, n_channels) for 2D covariance.
        cov_n: np.ndarray
            Covariance matrix of the nuisance maps.
            Shape should match cov.
        input_shapes: tuple
            Shapes of the input maps.
        compsep_run: dict   
            Dictionary with component separation parameters. See `gpilc` function for details.
        depro_cmb: float, optional
            Deprojection factor for CMB. If None, no deprojection is applied.
            Otherwise CMB residuals in GPILC maps will be at the level of depro_cmb * CMB_input.  
        m_bias: int or float, optional
            It will include m_bias more (if m_bias > 0) or less (if m_bias < 0) modes in the reconstructed GPILC maps.
    
    Returns
    -------
        np.ndarray
            Weights for GPILC. Shape is (n_pixels, n_channels, n_channels) for 3D covariance or (n_channels, n_channels) for 2D covariance.
    """  

    bandwidths = _get_bandwidths(config, compsep_run["good_channels"])
    A_cmb = _get_CMB_SED(np.array(config.instrument.frequency)[compsep_run["good_channels"]], 
                units=config.units, bandwidths=bandwidths) 

    if cov.ndim == 2:
        inv_cov = lg.inv(cov)

        m = _get_gilc_m(λ)
        m += int(m_bias)  

        U_s = np.delete(U,np.where(λ < λ[m-1]),axis=1)

        F = scipy.linalg.sqrtm(cov_n) @ U_s
        
        if depro_cmb is None:
            W = F @ lg.inv(F.T @ inv_cov @ F) @ F.T @ inv_cov
        else:
#            F_e = np.column_stack([F, depro_cmb * np.ones(input_shapes[0])])
            F_e = np.column_stack([F, depro_cmb * A_cmb])
            F_A = np.column_stack([F, A_cmb])
            W =  F_e @ lg.inv(F_A.T @ inv_cov @ F_A) @ F_A.T @ inv_cov

    elif cov.ndim == 3:
        m = _get_gilc_m(λ)
        m += int(m_bias)  

        covn_sqrt = lg.cholesky(cov_n)
        covn_sqrt_inv = lg.inv(covn_sqrt)

        W_=np.zeros((cov.shape[0],input_shapes[0],input_shapes[0]))

        for m_ in np.unique(m):
            U_s = U[m==m_,:,:m_]
            cov_inv = lg.inv(cov[m==m_])
            F = np.einsum("kij,kjz->kiz", covn_sqrt[m==m_],U_s)

            if depro_cmb is None: 
                W_[m==m_] = np.einsum("kil,klj->kij",F,np.einsum("kzl,klj->kzj",lg.inv(np.einsum("kiz,kij,kjl->kzl",F,cov_inv,F)),np.einsum("kiz,kij->kzj",F,cov_inv)))
            else:
#                e_cmb = depro_cmb * np.ones((F.shape[0],F.shape[1],1)) 
#                F_e = np.concatenate((F,e_cmb),axis=2)
                F_e = np.concatenate((F,np.tile(depro_cmb * A_cmb, (F.shape[0], 1))[:, :, np.newaxis]),axis=2)
                F_A = np.concatenate((F,np.tile(A_cmb, (F.shape[0], 1))[:, :, np.newaxis]),axis=2)
                W_[m==m_] = np.einsum("kil,klj->kij",F_e,np.einsum("kzl,klj->kzj",lg.inv(np.einsum("kiz,kij,kjl->kzl",F_A,cov_inv,F_A)),np.einsum("kiz,kij->kzj",F_A,cov_inv)))

        if "mask" not in compsep_run and W_.shape[0] != input_shapes[-2]:
            W = np.zeros((input_shapes[-2],W_.shape[1],W_.shape[2]))
            for i, k in np.ndindex(W_.shape[1],W_.shape[2]):
                W[:,i,k]=hp.ud_grade(W_[:,i,k],hp.npix2nside(input_shapes[-2]))
        else:
            W=np.copy(W_)

    return W

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]
                    





import numpy as np
import healpy as hp
import sys
#from sklearn.linear_model import LinearRegression

def purify_master(QU_maps,mask,lmax, return_E=True, return_B=True, purify_E=False):
    """
    Purify the E and B modes of the input QU maps using purification in NaMaster package.

    Parameters
    ----------
        QU_maps : array of Healpix maps
            The input maps containing Q and U components.
        mask : array
            The mask to apply to the maps.
        lmax : int
            The maximum multipole to consider for purification.
        return_E : bool, optional
            If True, return the E-mode alms. Default is True.
        return_B : bool, optional
            If True, return the B-mode alms. Default is True.
        purify_E : bool, optional
            If True, perform purification of E modes. Default is False.
    
    Returns
    -------
        alms_p : array
            The purified alms for E and/or B modes based on the input flags.
    """
        
#    try:
#        import pymaster as nmt
#    except ImportError:
#        raise ImportError("pymaster is required for NaMaster purification. Please install it.")
    from .spectra import nmt
#    print('Performing NaMaster purification.')

    maskbin = np.zeros_like(mask)
    maskbin[mask > 0.] = 1.
    nside = hp.get_nside(QU_maps[0])
    
    fp = nmt.NmtField(mask, [(QU_maps[0])*maskbin, (QU_maps[1])*maskbin],lmax=lmax,lmax_mask=lmax)
    alms_p, _ = fp._purify(fp.mask, fp.get_mask_alms(), [(QU_maps[0])*maskbin, (QU_maps[1])*maskbin], n_iter=fp.n_iter,task=[purify_E,True])
    if return_E and return_B:
        return alms_p
    elif return_E and not return_B:
        return alms_p[0]
    elif return_B and not return_E:
        return alms_p[1]
    else:
        raise ValueError("At least one of 'return_E' and 'return_B' must be True.")

def purify_recycling(QU_maps, QU_full_maps, mask,lmax, return_E=True, return_B=True, purify_E=False, iterations=0, **kwargs):
    """
    Purify the E and B modes of the input QU maps using recycling technique.

    Parameters
    ----------
        QU_maps : array of Healpix maps
            The input maps containing Q and U components.
        QU_full_maps : array of Healpix maps
            The Q and U maps containing the full signal over which the purification is actually performed.
        mask : array
            The mask to apply to the maps.
        lmax : int
            The maximum multipole to consider for purification.
        return_E : bool, optional
            If True, return the E-mode alms. Default is True.
        return_B : bool, optional
            If True, return the B-mode alms. Default is True.
        purify_E : bool, optional
            If True, perform purification of E modes. Default is False.
        iterations : int, optional
            Number of iterations for recycling purification. Default is 0.
    
    Returns
    -------
        alms_p : array
            The purified alms for E and/or B modes based on the input flags.
    """
#    print('Performing recycling purification.')
    maskbin = np.zeros_like(mask)
    maskbin[mask > 0.] = 1.
    lm = hp.Alm.getsize(lmax)

    nside = hp.get_nside(QU_maps[0])

    if return_E and return_B:
        alms_p = np.zeros((2,lm),dtype=complex)
    elif (return_E and not return_B) or (return_B and not return_E):
        alms_p = np.zeros((1,lm),dtype=complex)
    else:
        raise ValueError("At least one of 'return_E' and 'return_B' must be True.")
    
    TQU_maps = np.array([0.*QU_maps[0],QU_maps[0],QU_maps[1]])
    TQU_full_maps = np.array([0.*QU_full_maps[0],QU_full_maps[0],QU_full_maps[1]])

    if return_E and not purify_E:
        alms_p[0] = hp.map2alm(TQU_maps*mask, lmax=lmax, pol=True, **kwargs)[1]
    elif return_E and purify_E:
        print("Purification of E not implemented yet for recycling technique.")

    if return_B:
        alms_m = hp.map2alm(TQU_maps*maskbin, lmax=lmax, pol=True, **kwargs)
        full_alms_m = hp.map2alm(TQU_full_maps*maskbin, lmax=lmax, pol=True, **kwargs)
            
        alms_E = np.zeros((3,alms_m.shape[1]),dtype=complex)
        alms_B = np.zeros((3,alms_m.shape[1]),dtype=complex)
        alms_E[1] = alms_m[1]
        alms_B[2] = alms_m[2]
        full_alms_E = np.zeros((3,alms_m.shape[1]),dtype=complex)
        full_alms_B = np.zeros((3,alms_m.shape[1]),dtype=complex)
        full_alms_E[1] = full_alms_m[1]
        full_alms_B[2] = full_alms_m[2]

        maps_TQU_E=hp.alm2map(alms_E, nside, lmax=lmax, pol=True)
        maps_TQU_B=hp.alm2map(alms_B, nside, lmax=lmax, pol=True)
        full_maps_TQU_E=hp.alm2map(full_alms_E, nside, lmax=lmax, pol=True)
        full_maps_TQU_B=hp.alm2map(full_alms_B, nside, lmax=lmax, pol=True)

        alms_E_B=hp.map2alm(maps_TQU_E*maskbin, lmax=lmax, pol=True, **kwargs)[2]
        full_alms_E_B=hp.map2alm(full_maps_TQU_E*maskbin, lmax=lmax, pol=True, **kwargs)[2]

        alms_B_m = np.zeros((3,alms_m.shape[1]),dtype=complex)
        alms_B_m[2]=alms_E_B
        maps_TQU_B_temp = hp.alm2map(alms_B_m, nside, lmax=lmax, pol=True)
        full_alms_B_m = np.zeros((3,full_alms_m.shape[1]),dtype=complex)
        full_alms_B_m[2]=full_alms_E_B
        full_maps_TQU_B_temp = hp.alm2map(full_alms_B_m, nside, lmax=lmax, pol=True)

        #reg_Q = LinearRegression(fit_intercept=False).fit(full_maps_TQU_B_temp[1,maskbin>0].reshape(-1, 1), (full_maps_TQU_B)[1,maskbin>0])
        #reg_U = LinearRegression(fit_intercept=False).fit(full_maps_TQU_B_temp[2,maskbin>0].reshape(-1, 1), (full_maps_TQU_B)[2,maskbin>0])
        reg_Q, _, _, _ = np.linalg.lstsq(full_maps_TQU_B_temp[1,maskbin>0].reshape(-1, 1), 
                (full_maps_TQU_B)[1,maskbin>0], rcond=None)
        reg_U, _, _, _ = np.linalg.lstsq(full_maps_TQU_B_temp[2,maskbin>0].reshape(-1, 1),
                (full_maps_TQU_B)[2,maskbin>0], rcond=None)

        QU_p = np.zeros((3,12*nside**2))
#        QU_p[1] = maps_TQU_B[1]-((reg_Q.coef_)[0])*maps_TQU_B_temp[1]
#        QU_p[2] = maps_TQU_B[2]-((reg_U.coef_)[0])*maps_TQU_B_temp[2]
        QU_p[1] = maps_TQU_B[1]-(reg_Q[0])*maps_TQU_B_temp[1]
        QU_p[2] = maps_TQU_B[2]-(reg_U[0])*maps_TQU_B_temp[2]
            
        if iterations > 0:
            for it in range(iterations):
                alms_B_m = np.zeros((3,alms_m.shape[1]),dtype=complex)
                alms_B_m[2] = hp.map2alm([0.*(QU_p[0]),(QU_p[1])*maskbin,(QU_p[2])*maskbin], lmax=lmax, pol=True, **kwargs)[2]
                QU_p = hp.alm2map(alms_B_m,nside,lmax=lmax,pol=True)
            
        if return_E:
            alms_p[1] = hp.map2alm([0.*(QU_p[0]),(QU_p[1])*mask,(QU_p[2])*mask], lmax=lmax, pol=True, **kwargs)[2]
        else:
            alms_p[0] = hp.map2alm([0.*(QU_p[0]),(QU_p[1])*mask,(QU_p[2])*mask], lmax=lmax, pol=True, **kwargs)[2]
    
    if (return_E and not return_B) or (return_B and not return_E):
        return alms_p[0]
    else:
        return alms_p

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]


    


import numpy as np
import healpy as hp
import fnmatch
from types import SimpleNamespace
from .spectra import nmt
import os
import sys

from .routines import _slice_outputs, obj_out_to_array, _slice_data
from .configurations import Configs

REMOTE = 'https://irsa.ipac.caltech.edu/data/Planck/release_2/'
import os.path as op
from astropy.utils.data import download_file
import astropy
from typing import Any, Dict, Optional, Union, Tuple


def _preprocess_mask(mask: np.ndarray, nside_out: int) -> np.ndarray:
    """
    Preprocess a HEALPix mask by adjusting its resolution and assessing if it is binary or hits map.

    Parameters
    ----------
        mask : np.ndarray
            The input HEALPix mask, assumed to be a 1D array.
        nside_out : int
            Desired output NSIDE resolution.

    Returns
    -------
        np.ndarray
            The resampled HEALPix mask.

    Raises
    ------
        ValueError
            If the input is not a valid HEALPix mask.
    """
    if isinstance(mask, np.ndarray):
        try:
            nside_mask = hp.get_nside(mask)
#            if not is_binary_mask(mask):
#                print("Provided mask is not binary. Mask is assumed to be a hits map.")
            if nside_mask < nside_out:
                print("Provided mask has lower HEALPix resolution than that required for outputs. Mask will be upgraded to the output resolution.")
                mask = _upgrade_mask(mask, nside_out)
            elif nside_mask > nside_out:
                print("Provided mask has higher HEALPix resolution than that required for outputs. Mask will be downgraded to the output resolution.")
                mask = _downgrade_mask(mask, nside_out, threshold=0.5)
        except:
            raise ValueError("Invalid mask. It must be a valid HEALPix mask.")
        return mask
    else:
        raise ValueError("Invalid mask. It must be a numpy array.")

def _upgrade_mask(mask: np.ndarray, nside_out: int) -> np.ndarray:
    """
    Upgrade a mask to a higher HEALPix resolution.

    Parameters
    ----------
        mask : np.ndarray
            The input mask.
        nside_out : int
            The desired NSIDE.

    Returns
    -------
        np.ndarray
            The upgraded mask.
    """

    return hp.ud_grade(mask, nside_out, power=None if is_binary_mask(mask) else -2)


def _downgrade_mask(mask: np.ndarray, nside_out: int, threshold: float = 0.5) -> np.ndarray:
    """
    Downgrade a mask to a lower HEALPix resolution.

    Parameters
    ----------
        mask : np.ndarray
            The input mask.
        nside_out : int
            The desired NSIDE.
        threshold : float, optional
            Threshold for binary conversion after downgrading.

    Returns
    -------
        np.ndarray
            The downgraded (possibly binarized) mask.
    """
    if is_binary_mask(mask):
        mask = hp.ud_grade(mask, nside_out)
        return (mask > threshold).astype(float)
    else:
        return hp.ud_grade(mask, nside_out, power=-2)

def is_binary_mask(mask: np.ndarray) -> bool:
    """
    Check if a mask is binary (contains only 0.0 and 1.0).

    Parameters
    ----------
        mask : np.ndarray
            The input mask.

    Returns
    -------
        bool
            True if binary, False otherwise.
    """
    return np.all(np.isin(mask, [0., 1.]))

def _get_mask(config: Configs, compute_cls: Dict[str, Any], nsim: Optional[str] = None) -> np.ndarray:
    """
    Determine and return the appropriate mask based on configuration and compute_cls settings.

    Parameters
    ----------
        config: Configs
            Global configuration object. It specifies:
                - mask_covariance: Path to the mask fits file which has been used in component separation. Optional. 
                            It is used only if compute_cls["mask_type"] is None.
                - nside: HEALPix resolution of the component separation products.
                - fwhm_out: Full width at half maximum of output maps.
                - lmax: Maximum multipole for the outputs and power spectra computation.   
        compute_cls: Dict[str, Any])
            Dictionary with masking and output details. It should contain:
                - mask_type: String defining the type of mask to be used (e.g., 'from_fits', 'GAL*+fgres', etc.).
                - mask_path: String defining the path to the mask file if mask_type is 'from_fits'. 
                - outputs: Object containing the outputs of the component separation.
                - path: Path to the directory where the outputs are stored.
                - field_out: String defining the field associated with the output maps (e.g., 'TQU', 'QU', etc.) loaded from path.
                - field_cls_in: String defining the fields associated with the input maps (e.g., 'TQU', 'QU', etc.) used to compute output spectra.
                                It can be a slicing of field_out.
                - fsky (Optional[float]): Final sky fraction of the mask, used if 'fgres' is in mask_type.
                - smooth_tracer (Optional[float]): Smoothing scale for the tracer used in thresholding the mask. Used if 'fgres' is in mask_type.

        nsim: str, optional
            Simulation index, used to load simulation-specific data.

    Returns
    -------
        np.ndarray
            Final mask array, potentially multi-field (shape: [n_fields, npix]).
    """
    # Derive the number of fields and pixels from the outputs
    n_fields_in = obj_out_to_array(compute_cls["outputs"]).shape[-2] if obj_out_to_array(compute_cls["outputs"]).ndim == 3 else 1
    npix = obj_out_to_array(compute_cls["outputs"]).shape[-1]
    
    # Define the mask patterns to match against compute_cls["mask_type"]
    mask_patterns = ['GAL*+fgres', 'GAL*+fgtemp', 'GAL*+fgtemp^3','GAL*0', 'GAL97', 'GAL99', 'fgres', 'fgtemp', 
            'fgtemp^3', 'config+fgres', 'config+fgtemp', 'config+fgtemp^3', 'config']

    # Case 1: No mask defined
#    if compute_cls["mask_type"] is None and config.mask_path is None:
    if compute_cls["mask_type"] is None and config.mask_observations is None and config.mask_covariance is None:
        return np.ones(npix) if n_fields_in == 1 else np.ones((n_fields_in, npix))

    # Case 2: Use config-defined mask
#    if compute_cls["mask_type"] is None and config.mask_path is not None:
#        mask_spectra = _preprocess_mask(hp.read_map(config.mask_path, field=0), config.nside)
    if compute_cls["mask_type"] is None and (config.mask_observations is not None or config.mask_covariance is not None):
        _, mask_spectra = get_masks_for_compsep(config.mask_observations, config.mask_covariance, config.nside)
        return mask_spectra if n_fields_in == 1 else np.repeat(mask_spectra[np.newaxis, :], n_fields_in, axis=0)
        
    # Case 3: Load from FITS file directly        
    if compute_cls["mask_type"] == "from_fits":
        if "mask_path" not in compute_cls or not isinstance(compute_cls["mask_path"], str):
            raise ValueError("mask_path must be a string and defined in compute_cls when mask_type is 'from_fits'")

        if n_fields_in == 1:
            return _preprocess_mask(hp.read_map(compute_cls["mask_path"], field=0), config.nside)
        
        mask_spectra = np.zeros((n_fields_in, npix))
        mask_spectra[0] = _preprocess_mask(hp.read_map(compute_cls["mask_path"], field=0), config.nside)
        if compute_cls["field_cls_in"] == "TQU":
            try:
                mask_spectra[1] = _preprocess_mask(hp.read_map(compute_cls["mask_path"], field=1), config.nside)
                mask_spectra[2] = _preprocess_mask(hp.read_map(compute_cls["mask_path"], field=1), config.nside)
            except IndexError:
                mask_spectra[1:3] = mask_spectra[0]
        elif compute_cls["field_cls_in"] in ["QU","QU_E","QU_B"]:
            mask_spectra[1] = mask_spectra[0]
        else:
            for i in range(1,n_fields_in):
                try:
                    mask_spectra[i] = _preprocess_mask(hp.read_map(compute_cls["mask_path"], field=i), config.nside)
                except IndexError:
                    mask_spectra[i] = mask_spectra[0]
        return mask_spectra

    # Case 4: Named pattern (e.g., GAL99+fgres, config, etc.)
    if any(fnmatch.fnmatch(compute_cls["mask_type"], pattern) for pattern in mask_patterns):
        # Getting the initial mask
        if 'GAL' in compute_cls["mask_type"]:
            gal_masks_list = ['GAL20','GAL40','GAL60','GAL70','GAL80','GAL90','GAL97','GAL99']
            if compute_cls["mask_type"][:5] not in gal_masks_list:
                raise ValueError("GAL mask must be one of {}".format(", ".join(gal_masks_list)))
            idx_m = gal_masks_list.index(compute_cls["mask_type"][:5])
            rot = hp.Rotator(coord=f"G{config.coordinates}") if config.coordinates != "G" else None
            mask_init = hp.ud_grade(get_planck_mask(0, field=idx_m, nside=512),hp.npix2nside(npix))
            # Applying coordinate rotation if needed
            if config.coordinates != "G":
                alm_mask = hp.map2alm(mask_init, lmax=2*hp.npix2nside(npix), pol=False)
                rot.rotate_alm(alm_mask, inplace=True)
                mask_init = hp.alm2map(alm_mask, nside_out=hp.npix2nside(npix), lmax=2*hp.npix2nside(npix), pol=False)
            mask_init = mask_init == 1.
        elif 'config' in compute_cls["mask_type"]:
#            mask_init = np.ones(npix) if config.mask_path is None else _preprocess_mask(hp.read_map(config.mask_path, field=0), config.nside)
            if config.mask_observations is not None or config.mask_covariance is not None:
                _, mask_init = get_masks_for_compsep(config.mask_observations, config.mask_covariance, config.nside)
            else:
                mask_init = np.ones(npix)
        else:
            mask_init = np.ones(npix)

        # Generating and returning the final mask according to the mask_type
        if 'fgres' in compute_cls["mask_type"] or 'fgtemp' in compute_cls["mask_type"]:
            mask_init = mask_init == 1.

            if not "fsky" in compute_cls:
                raise ValueError("fsky must be defined in compute_cls when using 'fgres' or 'fgtemp' in mask_type")
            
            if 'fgres' in compute_cls["mask_type"]:
                if not hasattr(compute_cls["outputs"], 'fgds_residuals'):
                    from .compsep import _load_outputs_
                    fgres = SimpleNamespace()
                    filename = os.path.join(
                        compute_cls["path"],
                        f"fgds_residuals/{compute_cls['field_out']}_fgds_residuals_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}"
                    )
                    fgres.total = _load_outputs_(filename, compute_cls["field_out"], nsim=nsim)
                    if len(compute_cls['field_out']) > 1:
                        #fgres = _slice_data(fgres, compute_cls["field_out"], compute_cls["field_cls_in"])
                        fgres = _slice_outputs(fgres,compute_cls["field_out"],compute_cls["field_cls_in"])
                    return get_threshold_mask(fgres.total,mask_init,compute_cls["field_cls_in"],compute_cls["fsky"],config.lmax,smooth_tracer=compute_cls["smooth_tracer"])
                else:
                    return get_threshold_mask(compute_cls["outputs"].fgds_residuals, mask_init, compute_cls["field_cls_in"], compute_cls["fsky"], config.lmax, smooth_tracer=compute_cls["smooth_tracer"])
            
            else:
                if 'fgres_temp_for_masking' not in compute_cls:
                    raise ValueError("Template of fgds residuals must be defined in compute_cls as 'fgres_temp_for_masking' when using 'fgtemp' in mask_type")
                
                from .compsep import _load_outputs_
                fgres = SimpleNamespace()
                filename = os.path.join(
                    compute_cls["path"],
                    f"fgres_templates/{compute_cls['fgres_temp_for_masking']}/{compute_cls['field_out']}_fgres_templates_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}"
                )
                fgres.total = _load_outputs_(filename, compute_cls["field_out"], nsim=nsim)
                if len(compute_cls['field_out']) > 1:
                    fgres =  _slice_outputs(fgres,compute_cls["field_out"],compute_cls["field_cls_in"])
                if 'fgtemp^3' in compute_cls["mask_type"]:
                    return get_threshold_mask(fgres.total**3,mask_init,compute_cls["field_cls_in"],compute_cls["fsky"],config.lmax,smooth_tracer=compute_cls["smooth_tracer"])
                else:
                    return get_threshold_mask(fgres.total,mask_init,compute_cls["field_cls_in"],compute_cls["fsky"],config.lmax,smooth_tracer=compute_cls["smooth_tracer"])
                                    
        else:
            return mask_init if n_fields_in == 1 else np.repeat(mask_init[np.newaxis, :], n_fields_in, axis=0)

    raise ValueError("Mask not defined. Please check 'mask_type' and 'mask_path' in compute_cls.")

def get_threshold_mask(
    map_: np.ndarray,
    mask_init: np.ndarray,
    field_cls_in: str,
    fsky: float,
    lmax: int,
    smooth_tracer: float = 3.0
) -> np.ndarray:
    """
    Apply a threshold-based masking strategy to reduce the sky fraction.

    Parameters
    ----------
        map_: np.ndarray
            Input residual tracer map(s), shape (n_fields, npix) or (npix,) to be thresholded.
        mask_init: np.ndarray
            Initial binary mask.
        field_cls_in: str
            Fields associated to map_ (e.g., "TQU", "QU").
        fsky: float
            Desired final sky fraction.
        lmax: int
            Maximum multipole to consider.
        smooth_tracer: float
            FWHM in degrees for smoothing the tracer map.

    Returns
    ----------
        np.ndarray
            Final binary mask with the desired sky fraction.
    """
    fsky_in = np.mean(mask_init**2)

    # If the initial sky fraction is already lower than the desired one, return the initial mask
    if fsky > fsky_in:
        return mask_init if map_.ndim == 1 else np.repeat(mask_init[np.newaxis, :], map_.shape[0], axis=0)

    npix = map_.shape[-1]
    npix_mask = int((fsky_in - fsky) * npix)

    # If the input map is a scalar field, apply the thresholding and return the mask
    if map_.ndim == 1:
        return threshold_scalar_tracer(map_, mask_init, npix_mask, lmax, smooth_tracer=smooth_tracer)

    mask_spectra = np.ones((map_.shape[0], npix))
    # If input tracers include polarization spin-2 fields, corresponding mask will be generated from thresholding P tracer.
    # If input tracers are scalar fields, corresponding mask will be generated from thresholding scalar tracers.
    if field_cls_in == "TQU":
        mask_spectra[0] = threshold_scalar_tracer(map_[0],mask_init,npix_mask,lmax,smooth_tracer=smooth_tracer)
        mask_spectra_P = threshold_P_tracer(map_[1:],mask_init,npix_mask,lmax,smooth_tracer=smooth_tracer)
        mask_spectra[1:3] = mask_spectra_P
    elif field_cls_in == "QU":
        mask_spectra_P = threshold_P_tracer(map_,mask_init,npix_mask,lmax,smooth_tracer=smooth_tracer)
        mask_spectra[:2] = mask_spectra_P
    else:
        for i in range(map_.shape[0]):
            mask_spectra[i] = threshold_scalar_tracer(map_[i],mask_init,npix_mask,lmax,smooth_tracer=smooth_tracer)

    return mask_spectra

def _smooth_masks(mask: np.ndarray, apodization: str, smooth_mask: float) -> np.ndarray:
    """
    Apply smoothing/apodization to a given mask or set of masks.

    Parameters
    ----------
        mask: np.ndarray
            Input binary mask(s).
        apodization: str
            Apodization type ("gaussian", "gaussian_nmt", "C1", "C2").
        smooth_mask: float
            Smoothing scale in degrees.

    Returns
    ----------
        np.ndarray
            Smoothed mask.
    """
    if mask.ndim == 1:
        return _smooth_mask(mask, apodization, smooth_mask)

    return np.array([_smooth_mask(m, apodization, smooth_mask) for m in mask])

def _smooth_mask(mask_in: np.ndarray, apodization: str, smooth_mask: float) -> np.ndarray:
    """
    Smooth a single mask using the specified apodization technique.

    Parameters
    ----------
        mask_in: np.ndarray
            Input binary mask.
        apodization: str
            Apodization type ("gaussian", "gaussian_nmt", "C1", "C2").
        smooth_mask: float
            Smoothing scale in degrees.

    Returns
    ----------
        np.ndarray
            Smoothed mask.
    """
    if apodization == "gaussian":
        return hp.smoothing(mask_in,fwhm=np.radians(smooth_mask))
    if apodization == "gaussian_nmt":
        return nmt.mask_apodization(mask_in, smooth_mask, apotype="Smooth")
    if apodization in ["C1", "C2"]:
        return nmt.mask_apodization(mask_in, smooth_mask, apotype=apodization)
    raise ValueError('apodization must be either "gaussian", "gaussian_nmt", "C1" or "C2"')

def threshold_scalar_tracer(
    map_: np.ndarray,
    mask_in: np.ndarray,
    npix_mask: int,
    lmax: int,
    smooth_tracer: float = 3.0
) -> np.ndarray:
    """
    Generate a threshold mask from a scalar tracer map.

    Parameters
    ----------
        map_: np.ndarray
            Scalar tracer map.
        mask_in: np.ndarray
            Initial mask.
        npix_mask: int
            Number of additional pixels to mask.
        lmax: int
            Maximum multipole to be considerd for smoothing the tracer map.
        smooth_tracer: float
            Smoothing FWHM in degrees.

    Returns
    ----------
        np.ndarray
            Threshold mask.
    """

    smoothed_tracer = hp.smoothing(np.absolute(map_),fwhm=np.radians(smooth_tracer),lmax=lmax,pol=False)
    #smoothed_tracer = np.absolute(hp.smoothing(map_,fwhm=np.radians(smooth_tracer),lmax=lmax,pol=False))

    mask_spectra = np.ones_like(map_)
    idx_mask = np.argsort(smoothed_tracer * mask_in)[-npix_mask:]
    mask_spectra[idx_mask] = 0.
    mask_spectra[mask_in == 0.] = 0.
    return mask_spectra

def threshold_P_tracer(
    map_: np.ndarray,
    mask_in: np.ndarray,
    npix_mask: int,
    lmax: int,
    smooth_tracer: float = 3.0
) -> np.ndarray:
    """
    Generate a threshold mask from polarization tracers maps (Q, U).

    Parameters
    ----------
        map_: np.ndarray
            Polarization tracers maps [Q, U]. Shape: (2, npix).
        mask_in: np.ndarray
            Initial mask.
        npix_mask: int
            Number of pixels to mask.
        lmax: int
            Maximum multipole to be considerd for smoothing the tracers maps.
        smooth_tracer: float
            Smoothing FWHM in degrees.

    Returns
    ----------
        np.ndarray: 
            Thresholded polarization mask.
    """

    mask_spectra = np.ones_like(map_[0])
    map_ = hp.smoothing([0.*map_,map_[0],map_[1]],fwhm=np.radians(smooth_tracer),lmax=lmax,pol=True)[1:]
    map_P = np.sqrt((map_[0])**2 + (map_[1])**2)
    idx_mask = np.argsort(map_P * mask_in)[-npix_mask:]  #
    mask_spectra[idx_mask] = 0.
    mask_spectra[mask_in == 0.] = 0.
    return mask_spectra

def _get_processed_dir():
    """
    Returns the path to the processed Planck directory in the astropy cache,
    creating it if necessary.

    Returns
    -------
        str: 
            Directory path.
    """
    processed_dir = op.join(astropy.config.get_cache_dir(),
                            'processed', 'planck')
    if not op.exists(processed_dir):
        os.makedirs(processed_dir)

    return processed_dir

def get_planck_mask(apo: int = 5, nside: int = 2048, field: int = 3, info: bool = False) -> Union[np.ndarray, tuple]:
    """
    Downloads and processes a Planck Galactic mask.

    Parameters
    -------    
        apo: int 
            Apodization level (default 5).
        nside: int 
            HEALPix resolution.
        field: int 
            Field to read.
        info: bool 
            Whether to return header info.

    Returns
    -------
        np.ndarray or tuple: 
            Mask data or (mask, header).
    """
    remote_url = op.join(REMOTE, 'ancillary-data/masks',
                         f'HFI_Mask_GalPlane-apo{apo}_2048_R2.00.fits')
    remote_file = download_file(remote_url, cache=True)

    if info:
        return hp.read_map(remote_file, field=[], h=True)

    local_file = op.join(_get_processed_dir(),
                         f'HFI_Mask_GalPlane-apo{apo}_{nside}_R2.00.fits')
    try:
        return hp.read_map(local_file, field=field)
    except IOError:
        output = hp.read_map(remote_file, field=None)
        output = hp.ud_grade(output, nside)
        hp.write_map(local_file, output)
        return output[field]

def get_masks_for_compsep(mask_obs: Union[str, np.ndarray, None]
    , mask_cov: Union[str, np.ndarray, None]
    , nside: int) -> Tuple[Union[np.ndarray, None], Union[np.ndarray, None]]:
    """
    Get observation and covariance masks for component separation.

    Parameters
    ----------
        mask_obs: Union[str, np.ndarray, None]
            Path to the observation mask fits file or a numpy array. If None, no observation mask is applied.
        mask_cov: Union[str, np.ndarray, None]
            Path to the covariance mask fits file or a numpy array. If None, the observation mask is used as the covariance mask.
        nside: int
            HEALPix resolution for the masks in output.
    
    Returns
    -------
        Tuple[Union[np.ndarray, None], Union[np.ndarray, None]]:
            Tuple containing the observation mask and covariance mask as numpy arrays. If no mask is provided, returns None.
    """

    if mask_obs is not None:
        if not isinstance(mask_obs, (str, np.ndarray)):
            raise ValueError("mask_observations must be a string full path to a HEALPix mask fits file or a numpy array.")

        if isinstance(mask_obs, str):
            mask_obs = hp.read_map(mask_obs, field=0)
        mask_obs = _preprocess_mask(mask_obs, nside)
        mask_obs /= np.max(mask_obs)  # Normalizing mask to have values between 0 and 1
    else:
        mask_obs = None
        
    if mask_cov is not None:
        if not isinstance(mask_cov, (str, np.ndarray)):
            raise ValueError("mask_covariance must be a string full path to a HEALPix mask fits file or a numpy array.")

        if isinstance(mask_cov, str):
            mask_cov = hp.read_map(mask_cov, field=0)
        mask_cov = _preprocess_mask(mask_cov, nside)
        mask_cov /= np.max(mask_cov)  # Normalizing mask to have values between 0 and 1
        if mask_obs is not None:
            mask_cov[mask_obs == 0.] = 0.  # Ensuring that the covariance mask is zero where the observation mask is zero
            mask_obs[mask_cov == 0.] = 0.  # Ensuring that the observation mask is zero where the covariance mask is zero
    else:
        mask_cov = mask_obs

    return mask_obs, mask_cov


__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]
                    



"""

# BROOM: Blind Reconstruction Of signals from Observations in the Microwaves

**BROOM** is a Python package for blind component separation and Cosmic Microwave Background (CMB) data analysis.

---

## 📦 Installation

You can install the base package using:

```
pip install cmbroom
```

This installs the core functionality.  
If you plan to use the few functions that depend on `pymaster`, **you must install it separately** (version `>=2.4`).

---

### 🔧 To include `pymaster` automatically:

You can install `cmbroom` along with its optional `pymaster` dependency by running:

```
pip install cmbroom[pymaster]
```

However, `pymaster` requires some additional system libraries to be installed **before** running the above command.

#### ✅ On Ubuntu/Debian:
```
sudo apt update
sudo apt install build-essential python3-dev libfftw3-dev libcfitsio-dev
```

#### ✅ On macOS (using Homebrew):
```
brew install fftw cfitsio
```

## Documentation

A detailed introduction to the parameters and simulation pipeline is available in:

- [**tutorials/tutorial_satellite.ipynb**](tutorials/tutorial_satellite.ipynb)   
- [**configs/config_demo.yaml**](broom/configs/config_demo.yaml) — Example configuration file

Component separation methods are covered in:

- [**tutorials/tutorial_satellite.ipynb**](tutorials/tutorial_satellite.ipynb) 
- [**tutorials/tutorial_satellite_part2.ipynb**](tutorials/tutorial_satellite_part2.ipynb) 

Power spectrum estimation is demonstrated in:

- [**tutorials/tutorial_spectra.ipynb**](tutorials/tutorial_spectra.ipynb)

For partial-sky, ground-based experiment analysis, see:

- [**tutorials/tutorial_groundbased.ipynb**](tutorials/tutorial_groundbased.ipynb) 

🔗 **Full online documentation:**  
👉 [https://alecarones.github.io/broom/](https://alecarones.github.io/broom/)


## References

Paper on **broom** package is in preparation.

If you use the following methodologies please cite the corresponding papers:

- ILC or NILC: [Bennett et al., 2003](https://arxiv.org/abs/astro-ph/0302207), [Delabrouille et al., 2009](https://arxiv.org/abs/0807.0773)
- cMILC: [Remazeilles et al., 2021](https://arxiv.org/abs/2006.08628), [Carones et al., 2024](https://arxiv.org/abs/2402.17579)
- MC-ILC or MC-NILC: [Carones et al., 2023](https://arxiv.org/abs/2212.04456)
- PILC: [Fernández-Cobos et al., 2016](https://arxiv.org/abs/1601.01515)
- cPILC: [Adak, 2021](https://arxiv.org/abs/2104.13778)
- GILC, GNILC, GPILC: [Remazeilles et al., 2011](https://arxiv.org/abs/1103.1166), [Planck Collaboration, 2016](https://arxiv.org/abs/1605.09387)
- foreground diagnostic: [Carones et al., 2024](https://arxiv.org/abs/2402.17579)
- power spectrum computation: [Gorski et al., 2005](https://arxiv.org/abs/astro-ph/0409513), [Zonca et al., 2019](https://ui.adsabs.harvard.edu/abs/2019JOSS....4.1298Z/abstract), [Alonso et al., 2019](https://arxiv.org/abs/1809.09603)

### 📦 Dependencies

This package relies on several scientific Python libraries:

- [astropy>=6.0.1](https://www.astropy.org/)
- [numpy>1.18.5](https://numpy.org/)
- [scipy>=1.8](https://scipy.org/)
- [healpy>=1.15](https://healpy.readthedocs.io/)
- [pysm3>=3.3.2](https://pysm3.readthedocs.io/en/latest/#)
- [mtneedlet>=0.0.5](https://javicarron.github.io/mtneedlet/)



"""


from .compsep import (
    component_separation, estimate_residuals, _combine_products, _load_outputs
)
from .simulations import _get_data_foregrounds_, _get_data_simulations_, _get_noise_simulation, _get_cmb_simulation, _get_full_simulations
from .configurations import Configs, get_params
from .spectra import _compute_spectra, _compute_spectra_, _load_cls
from .clusters import get_and_save_real_tracers_B


__version__ = "0.1.4"



import numpy as np
import healpy as hp
import sys
from .configurations import Configs
from .routines import _get_local_cov, _EB_to_QU, _E_to_QU, _B_to_QU, obj_to_array, array_to_obj, _log, _get_bandwidths
from .saving import _save_compsep_products, _get_full_path_out
from .needlets import _get_nside_lmax_from_b_ell, _get_needlet_windows_, _needlet_filtering, _get_good_channels_nl
from .ilcs import get_ilc_cov
from .seds import _get_CMB_SED
import scipy
from numpy import linalg as lg
from types import SimpleNamespace
import os
from typing import Any, Dict, Optional, Union, List, Tuple

def gilc(config: Configs, input_alms: SimpleNamespace, compsep_run: Dict[str, Any], **kwargs) -> Optional[SimpleNamespace]:
    """
    Performs Generalized Internal Linear Combination (GILC) component separation on scalar fields.

    Parameters
    ----------
        config: Configs
            Configuration object with general settings. It includes:
                - lmax : int, maximum multipole for the component separation.
                - nside : int, HEALPix resolution parameter for compsep products.
                - fwhm_out : float, full width at half maximum of the output beam in arcminutes.
                - pixel_window_out : bool, whether to apply a pixel window to the output maps.
                - field_out : str, desired output fields (e.g., "T", "E", "B", "QU", "TQU", "QU_E", "QU_B").
                - save_compsep_products : bool, whether to save component separation products.
                - return_compsep_products : bool, whether to return component separation products.
                - path_outputs : str, path to save the output files.
            
        input_alms: SimpleNamespace
            Input multifrequency alms of the scalar fields. 
            Each attribute should be a numpy array of shape (n_channels, (n_fields), n_alms, n_components), of the noise maps. 
            with n_fields > 1 if multiple scalars are provided.
            It should contain:
            - total : np.ndarray, alms of the total input maps.
            - nuisance : Optional[np.ndarray], alms of the nuisance maps (if available).
            Alternatively to nuisance, it can contain:
            - cmb : np.ndarray, alms of the CMB maps.
            - noise : np.ndarray, alms of the noise maps.
        
        compsep_run: Dict[str, Any]
            Dictionary with component separation parameters. It should include:
            - domain : str, either "pixel" or "needlet" for the component separation domain.
            - channels_out : list, indices of the frequency channels to reconstruct with GILC. Default is all channels.
            - depro_cmb : Optional[Union[float, list, np.ndarray]], deprojection factor for CMB (scalar or list per needlet bands). 
                    Default is None.
            - m_bias : Optional[Union[float, list, np.ndarray]], if not zero, it will include m_bias more (if m_bias > 0) 
                    or less (if m_bias < 0) modes in the reconstructed GNILC maps. Default is 0.
                    It can be a list if different values are needed for different needlet bands.
            - cmb_nuisance : bool, whether to include CMB alms in the nuisance covariance. Default is True.
            - needlet_config: Dictionary containing needlet settings. Needed if domain is "needlet". It should include:
                - "needlet_windows": Type of needlet windows ('cosine', 'standard', 'mexican').
                - "ell_peaks": List of integers defining multipoles of peaks for needlet bands (required for 'cosine').
                - "width": Width of the needlet windows (required for 'standard' and 'mexican').
                - "merging_needlets": Integer or list of integers defining ranges of needlets to be merged.
            - b_squared : bool, whether to square the needlet windows. Default is False.            
            - adapt_nside : bool, whether to adapt the nside based on the needlet windows. Default is False.
            - mask : Optional[np.ndarray], mask to apply to the maps (if available).
            - cov_noise_debias : Optional[Union[float, list, np.ndarray]],
                If not zero, it will debias the covariances by subtracting a term cov_noise_debias * noise_covariance.
                It can be a scalar or a list per needlet bands. Default is 0.
            
        **kwargs: 
            Dictionary of additional keyword arguments to pass to healpy function 'map2alm'.

    Returns
    -------
        Optional[SimpleNamespace]: 
            Output map object with reconstructed foreground maps if config.return_compsep_products is True.
    """

    compsep_run = _standardize_gnilc_run(compsep_run, input_alms.total.shape[0], config.lmax)
    
    compsep_run["nuis_idx"] = get_nuisance_idx(input_alms, compsep_run, config.verbose)
    if np.any(np.array(compsep_run["cov_noise_debias"]) != 0.):
        if not hasattr(input_alms, "noise"):
            raise ValueError("The input_alms object must have 'noise'' attribute for debiasing the covariance.")
        compsep_run["noise_idx"] = 2 if hasattr(input_alms, "fgds") else 1
#    print("Using nuisance alms with indices:", compsep_run["nuis_idx"])

    output_maps = _gilc(config, obj_to_array(input_alms), compsep_run, **kwargs)
    
    output_maps = _gilc_post_processing(config, output_maps, compsep_run, **kwargs)

    compsep_run.pop("nuis_idx", None)
    compsep_run.pop("noise_idx", None)

    outputs = array_to_obj(output_maps, input_alms)
    del output_maps

    if config.save_compsep_products:
        _save_compsep_products(config, outputs, compsep_run, nsim=compsep_run["nsim"])
    
    if config.return_compsep_products:
        return outputs
    return None

def fgd_diagnostic(config: Configs, input_alms: SimpleNamespace, compsep_run: Dict[str, Any], **kwargs) -> Optional[SimpleNamespace]:
    """
    Return diagnostic maps of the foreground complexity for the provided scalar fields.

    Parameters
    ----------
        config: Configs
            Configuration object with general settings. It includes:
                - lmax : int, maximum multipole for the component separation.
                - nside : int, HEALPix resolution parameter for compsep products.
                - fwhm_out : float, full width at half maximum of the output beam in arcminutes.
                - field_out : str, desired fields to be considered (e.g., "T", "E", "B", "TEB").
                - save_compsep_products : bool, whether to save diagnostic maps.
                - return_compsep_products : bool, whether to return diagnostic maps.
                - path_outputs : str, path to save the output files.
            
        input_alms: SimpleNamespace
            Input multifrequency alms of the scalar fields. 
            Each attribute should be a numpy array of shape (n_channels, (n_fields), n_alms, n_components), 
            with n_fields > 1 if multiple scalars are provided.
            It should contain:
            - total : np.ndarray, alms of the total input maps.
            - nuisance : Optional[np.ndarray], alms of the nuisance maps (if available).
            Alternatively to nuisance, it can contain:
            - cmb : np.ndarray, alms of the CMB maps.
            - noise : np.ndarray, alms of the noise maps.
    
        compsep_run: Dict[str, Any]
            Dictionary with diagnostic parameters. It should include:
                - domain : str, either "pixel" or "needlet" for the component separation domain.
                - cmb_nuisance : bool, whether to include CMB alms in the nuisance covariance. Default is True.
                - needlet_config: Dictionary containing needlet settings. Needed if domain is "needlet". It should include:
                    - "needlet_windows": Type of needlet windows ('cosine', 'standard', 'mexican').
                    - "ell_peaks": List of integers defining multipoles of peaks for needlet bands (required for 'cosine').
                    - "width": Width of the needlet windows (required for 'standard' and 'mexican').
                    - "merging_needlets": Integer or list of integers defining ranges of needlets to be merged.
                - b_squared : bool, whether to square the needlet windows. Default is False.            
                - adapt_nside : bool, whether to adapt the nside based on the needlet windows. Default is False.
                - mask : Optional[np.ndarray], mask to apply to the maps (if available).
            
        **kwargs: 
            Dictionary of additional keyword arguments to pass to healpy function 'map2alm'.

    Returns
    -------
        Optional[SimpleNamespace]
            Object containing foreground diagnostic maps.
    """

    if not "cmb_nuisance" in compsep_run:
        compsep_run["cmb_nuisance"] = True            
     
    compsep_run["nuis_idx"] = get_nuisance_idx(input_alms, compsep_run, config.verbose)
    if np.any(np.array(compsep_run["cov_noise_debias"]) != 0.):
        if not hasattr(input_alms, "noise"):
            raise ValueError("The input_alms object must have 'noise'' attribute for debiasing the covariance.")
        compsep_run["noise_idx"] = 2 if hasattr(input_alms, "fgds") else 1

    if isinstance(compsep_run["nuis_idx"], int):
        nuis_alms = (obj_to_array(input_alms))[...,compsep_run["nuis_idx"]]
    elif isinstance(compsep_run["nuis_idx"], list):
        nuis_alms = (obj_to_array(input_alms))[...,compsep_run["nuis_idx"][0]] + (obj_to_array(input_alms))[...,compsep_run["nuis_idx"][1]]
    inputs_alms_for_diagn = np.concatenate([
        input_alms.total[...,np.newaxis],
        nuis_alms[...,np.newaxis]],axis=-1)
    del nuis_alms

    if np.any(np.array(compsep_run["cov_noise_debias"]) != 0.):
        noi_alms = (obj_to_array(input_alms))[...,compsep_run["noise_idx"]]
        inputs_alms_for_diagn = np.concatenate([inputs_alms_for_diagn, noi_alms[...,np.newaxis]], axis=-1)
        del noi_alms

    output_maps = _fgd_diagnostic(config, inputs_alms_for_diagn, compsep_run)
    del inputs_alms_for_diagn

    compsep_run.pop("nuis_idx", None)
    compsep_run.pop("noise_idx", None)

    outputs = SimpleNamespace(m=output_maps)

    del output_maps

    if config.save_compsep_products:
        _save_compsep_products(config, outputs, compsep_run, nsim=compsep_run["nsim"])
    
    if config.return_compsep_products:
        return outputs
    return None

def get_nuisance_idx(
    input_alms: SimpleNamespace,
    compsep_run: Dict[str, Any],
    verbose: bool = False,
) -> Union[int, List[int]]:
    """
    Determines the index or indices of nuisance alms based on the input alms object and component separation parameters.

    Parameters
    ----------
        input_alms: SimpleNamespace
            Input alms object containing fields like 'total', 'nuisance', 'cmb', and 'noise'.
        compsep_run: Dict[str, Any]
            Component separation parameters including 'cmb_nuisance'.
        verbose: bool, optional
            If True, prints additional information about the nuisance alms being used. Default is False.
    
    Returns
    -------
        Union[int, List[int]]
            Index or list of indices for nuisance alms.
    
    Raises  
    ------
    ValueError
        If the input_alms object does not have the required attributes for nuisance covariance.
    """

    has_nuisance = hasattr(input_alms, "nuisance")
    has_noise    = hasattr(input_alms, "noise")
    has_fgds     = hasattr(input_alms, "fgds")
    has_cmb      = hasattr(input_alms, "cmb")

    # ――― Case A: explicit nuisance alms already present ―――
    if has_nuisance:
        if has_noise and has_fgds:
            return 3
        elif not has_noise and not has_fgds:
            return 1
        else:
            return 2
    
    # ――― Case B: need to build nuisance from CMB / noise ―――
    cmb_nuisance = compsep_run.get("cmb_nuisance", True)

    if cmb_nuisance and (not has_cmb or not has_noise):
        raise ValueError(
            "cmb_nuisance=True: 'cmb' and 'noise' must be present in input_alms."
        )
    if not cmb_nuisance and not has_noise:
        raise ValueError(
            "cmb_nuisance=False: 'noise' must be present in input_alms."
        )
    
    if cmb_nuisance:
        _log("No nuisance alms provided. Using input noise and CMB alms as nuisance.", verbose=verbose)
    else:
        _log("No nuisance alms provided. Using input noise alms as nuisance.", verbose=verbose)
        
    if not cmb_nuisance:
        # CMB excluded from nuisance covariance
        return 2 if has_fgds else 1
    else:
        # CMB included
        return [2, 3] if has_fgds else [1, 2]


def _gilc_post_processing(config: Configs, output_maps: np.ndarray, compsep_run: Dict[str, Any], **kwargs) -> np.ndarray:
    """
    Applies post-processing steps to GILC output maps, such as converting from E/B to Q/U or masking.

    Parameters
    ----------
        config: Configs
            Configuration object with general settings. See 'gilc' function for details.
        output_maps: np.ndarray
            Raw output maps from GILC. Shape must be (n_channels, (n_fields), npix, n_components).
        compsep_run: Dict[str, Any]
            Component separation parameters. See 'gilc' function for details.
        **kwargs: 
            Additional arguments for EB/QU conversions.

    Returns
    -------
        np.ndarray
            Processed output maps.
    """
    if output_maps.ndim == 4 and (
        (output_maps.shape[1] == 2 and config.field_out == "QU") or
        (output_maps.shape[1] == 3 and config.field_out == "TQU")
    ):
        outputs = np.zeros_like(output_maps)
        for f, c in np.ndindex(output_maps.shape[0],output_maps.shape[-1]):
            outputs[f,...,c] = _EB_to_QU(output_maps[f,...,c],config.lmax, **kwargs)

        if "mask" in compsep_run:
            outputs[:,:,compsep_run["mask"] == 0.,:] = 0.
        return outputs

    elif (output_maps.ndim==3) and (config.field_out in ["QU_E", "QU_B"]):
        output = np.zeros((output_maps.shape[0], 2, output_maps.shape[1], output_maps.shape[-1]))
        for f, c in np.ndindex(output_maps.shape[0],output_maps.shape[-1]):
            if config.field_out == "QU_E":
                output[f,...,c] = _E_to_QU(output_maps[f,:,c],config.lmax, **kwargs)
            elif config.field_out=="QU_B":
                output[f,...,c] = _B_to_QU(output_maps[f,:,c],config.lmax, **kwargs)
        if "mask" in compsep_run:
            output[:,:,compsep_run["mask"] == 0.,:] = 0.
        return output
        
    else:
        if "mask" in compsep_run:
            output_maps[...,compsep_run["mask"] == 0.,:] = 0.
        return output_maps

def _gilc(config: Configs, input_alms: np.ndarray, compsep_run: Dict[str, Any], **kwargs) -> np.ndarray:
    """
    Apply GILC component separation routine to each provided scalar field.

    Parameters
    ----------
        config: Configs
            Configuration object with general settings. See 'gilc' function for details.
        input_alms: np.ndarray
            Input alms (3D or 4D array). 
            Shape should be (n_channels, n_fields, n_alms, n_components) for 4D or (n_channels, n_alms, n_components) for 3D.
        compsep_run: Dict[str, Any]
            Parameters for the GILC component separation run. See 'gilc' function for details.
        **kwargs: 
            Additional keyword arguments for healpy function map2alm.

    Returns
    -------
        np.ndarray
            Output separated component maps. 
            Shape will be (n_channels_out, (n_fields), npix, n_components).
    """
    if input_alms.ndim == 4:
        output_maps = np.zeros((len(compsep_run["channels_out"]), input_alms.shape[1], 12 * config.nside**2, input_alms.shape[-1]))
        for i in range(input_alms.shape[1]):
            output_maps[:,i] = _gilc_scalar(config, input_alms[:, i], compsep_run, **kwargs)
    elif input_alms.ndim == 3:
        output_maps = _gilc_scalar(config, input_alms, compsep_run, **kwargs)
    return output_maps

def _fgd_diagnostic(config: Configs, input_alms: np.ndarray, compsep_run: Dict[str, Any]) -> np.ndarray:
    """
    Foreground diagnostic map generation.

    Parameters
    ----------
        config: Configs
            Configuration object with general settings. See 'fgd_diagnostic' function for details.
        input_alms: np.ndarray
            Input alms (4D or 3D array). 
            Shape should be (n_channels, (n_fields), n_alms, n_comps) being 4D if multiple fields are provided, 
            or 3D if only one field is provided.
        compsep_run: Dict[str, Any] 
            Parameters for foreground diagnostic. See 'fgd_diagnostic' function for details.

    Returns
    -------
        np.ndarray
            Output diagnostic maps.
            Shape will be ((n_fields), npix) if domain is "pixel", or ((n_fields), n_bands, npix) if domain is "needlet".
    """
    if input_alms.ndim == 4:
        if compsep_run["domain"]=="needlet":
            nls_number = _get_needlet_windows_(compsep_run["needlet_config"], config.lmax).shape[0]
            output_maps = np.zeros((input_alms.shape[1], nls_number, 12 * config.nside**2))
        else:
            output_maps = np.zeros((input_alms.shape[1], 12 * config.nside**2))

        for i in range(input_alms.shape[1]):
            output_maps[i] = _fgd_diagnostic_scalar(config, input_alms[:, i], compsep_run)

    elif input_alms.ndim == 3:
        output_maps = _fgd_diagnostic_scalar(config, input_alms, compsep_run)

    return output_maps

def _standardize_gnilc_run(compsep_run: Dict[str, Any], n_freqs: int, lmax: int) -> Dict[str, Any]:
    """
    Standardizes the `compsep_run` dictionary for GILC/FGD operations.

    Parameters
    ----------
        compsep_run : dict
            Dictionary of component separation configuration parameters. See `gilc` for details.
        n_freqs : int
            Number of frequency channels in the input data.
        lmax : int
            Maximum multipole to consider.

    Returns
    -------
        dict
            Standardized compsep_run dictionary.
    """
    if not "channels_out" in compsep_run:
        compsep_run["channels_out"] = list(range(n_freqs))
    else: 
        if np.any(np.array(compsep_run["channels_out"]) >= n_freqs):
            raise ValueError("Some of the requested channels_out are not present in the input maps.")

    if compsep_run["domain"]=="needlet":
        nls_number = _get_needlet_windows_(compsep_run["needlet_config"], lmax).shape[0]

    # CMB deprojection setup    
    if "depro_cmb" not in compsep_run or compsep_run["depro_cmb"] is None:
        compsep_run["depro_cmb"] = None if compsep_run["domain"] == "pixel" else np.repeat(None, nls_number)
    else:
        if compsep_run["domain"] == "pixel":
            if isinstance(compsep_run["depro_cmb"], (list, np.ndarray)):
                raise ValueError("depro_cmb must be a scalar or None when domain is pixel.")
        elif compsep_run["domain"] == "needlet":
            if isinstance(compsep_run["depro_cmb"], (int, float)):
                compsep_run["depro_cmb"] = np.repeat(compsep_run["depro_cmb"], nls_number)
            elif isinstance(compsep_run["depro_cmb"], list):
                compsep_run["depro_cmb"] = (compsep_run["depro_cmb"] + [None] * nls_number)[:nls_number]
            elif isinstance(compsep_run["depro_cmb"], np.ndarray):
                _len = nls_number - compsep_run["depro_cmb"].shape[0]
                if _len > 0:
                    compsep_run["depro_cmb"] = np.append(compsep_run["depro_cmb"], [None] * _len)
                compsep_run["depro_cmb"] = compsep_run["depro_cmb"][:nls_number]
            else:
                raise ValueError("depro_cmb must be a scalar, a list or a np.ndarray if domain is needlet.")

    # Bias setup
    if "m_bias" not in compsep_run or compsep_run["m_bias"] is None:
        compsep_run["m_bias"] = 0 if compsep_run["domain"] == "pixel" else np.repeat(0, nls_number)
    else:
        if compsep_run["domain"]=="pixel":
            if isinstance(compsep_run["m_bias"], (list, np.ndarray)):
                raise ValueError("m_bias must be a scalar when domain is pixel.")
        elif compsep_run["domain"]=="needlet":
            if isinstance(compsep_run["m_bias"], int):
                compsep_run["m_bias"] = np.repeat(compsep_run["m_bias"], nls_number)
            elif isinstance(compsep_run["m_bias"], list):
                compsep_run["m_bias"] = (compsep_run["m_bias"] + [0] * nls_number)[:nls_number]
            elif isinstance(compsep_run["m_bias"], np.ndarray):
                _len = nls_number - compsep_run["m_bias"].shape[0]
                if _len > 0:
                    compsep_run["m_bias"] = np.append(compsep_run["m_bias"], [0] * _len)
                compsep_run["m_bias"] = compsep_run["m_bias"][:nls_number]    
            else:
                raise ValueError("m_bias must be a scalar, a list or a np.ndarray if domain is needlet.")

    if "cmb_nuisance" not in compsep_run:
        compsep_run["cmb_nuisance"] = True            
     
    return compsep_run

def _gilc_scalar(config: Configs, input_alms: np.ndarray, compsep_run: Dict[str, Any], **kwargs) -> np.ndarray:
    """
    Apply GILC component separation in pixel or needlet domain.

    Parameters
    ----------
        config : Configs
            Configuration object. See `gilc` for details.
        input_alms : np.ndarray
            Input alm coefficients. Shape should be (n_channels, n_alms, n_components).
        compsep_run : dict
            Dictionary with GILC parameters. See `gilc` for details.

    Returns
    -------
        np.ndarray
            Separated component maps. Shape will be (n_channels_out, npix, n_components).
    """
    if compsep_run["domain"] == "pixel":
        return _gilc_pixel(config, input_alms, compsep_run, **kwargs)
    elif compsep_run["domain"] == "needlet":
        return _gilc_needlet(config, input_alms, compsep_run, **kwargs)
    else:
        raise ValueError(f"Unsupported domain: {compsep_run['domain']}")


def _fgd_diagnostic_scalar(config: Configs, input_alms: np.ndarray, compsep_run: Dict[str, Any]) -> np.ndarray:
    """
    Get Foreground Diagnostic maps for provided scalar field either in pixel or needlet domain.

    Parameters
    ----------
        config : Configs
            Configuration object. See `fgd_diagnostic` for details.
        input_alms : np.ndarray
            Input alm coefficients. Shape should be (n_channels, n_alms).
        compsep_run : dict
            Dictionary with run parameters. See `fgd_diagnostic` for details.

    Returns
    -------
        np.ndarray
            Foreground diagnostic maps. Shape will be (npix) if domain is "pixel", or (n_bands, npix) if domain is "needlet".
    """
    if compsep_run["domain"] == "pixel":
        return _fgd_diagnostic_pixel(config, input_alms, compsep_run)
    elif compsep_run["domain"] == "needlet":
        return _fgd_diagnostic_needlet(config, input_alms, compsep_run)
    else:
        raise ValueError(f"Unsupported domain: {compsep_run['domain']}")

def _gilc_pixel(config: Configs, input_alms: np.ndarray, compsep_run: Dict[str, Any], **kwargs) -> np.ndarray:
    """
    GILC in pixel domain.

    Parameters
    ----------
        config : Configs
            Configuration object. See `gilc` for details.
        input_alms : np.ndarray
            Input spherical harmonic coefficients. Shape should be (n_channels, n_alms, n_components).
        compsep_run : dict
            Component separation settings. See `gilc` for details.
        kwargs : dict
            Extra keyword arguments passed to alm/map conversion routines.

    Returns
    -------
        np.ndarray
            Output map with separated components. Shape will be (n_channels_out, npix, n_components).
    """

    compsep_run["good_channels"] = _get_good_channels_nl(config, np.ones(config.lmax+1))

    input_maps = np.zeros((compsep_run["good_channels"].shape[0], 12 * config.nside**2, input_alms.shape[-1]))
    for n, channel in enumerate(compsep_run["good_channels"]):
        input_maps[n] = np.array([hp.alm2map(np.ascontiguousarray(input_alms[channel, :, c]), config.nside, lmax=config.lmax, pol=False) for c in range(input_alms.shape[-1])]).T

    output_maps = _gilc_maps(
        config,
        input_maps,
        compsep_run,
        np.ones(config.lmax + 1),
        depro_cmb=compsep_run["depro_cmb"],
        m_bias=compsep_run["m_bias"],
        noise_debias=compsep_run["cov_noise_debias"]
    )
    
    del compsep_run['good_channels']

    if config.pixel_window_out:
        for f, c in np.ndindex(output_maps.shape[0],output_maps.shape[-1]):
            alm_out = hp.map2alm(output_maps[f,:,c],lmax=config.lmax, pol=False, **kwargs)
            output_maps[f,:,c] = hp.alm2map(alm_out, config.nside, lmax=config.lmax, pol=False, pixwin=True)

    if "mask" in compsep_run:
        output_maps[:,compsep_run["mask"] == 0.,:] = 0.

    return output_maps

def _fgd_diagnostic_pixel(config: Configs, input_alms: np.ndarray, compsep_run: dict) -> np.ndarray:
    """
    Generate pixel-domain diagnostic maps from input alms.

    Parameters
    ----------
        config: Configs
            Configuration object with general settings, including nside and lmax. See `fgd_diagnostic` for details.
        input_alms: np.ndarray
            Input spherical harmonic coefficients. Shape should be (n_channels, n_alms, n_comps).
        compsep_run: dict
            Dictionary with component separation parameters, including mask (if available). See `fgd_diagnostic` for details.


    Returns
    -------
        np.ndarray
            Diagnostic map of foregrounc complexity.
    """

    compsep_run["good_channels"] = _get_good_channels_nl(config, np.ones(config.lmax+1))

    input_maps = np.zeros((compsep_run["good_channels"].shape[0], 12 * config.nside**2, input_alms.shape[-1]))
    for n, channel in enumerate(compsep_run["good_channels"]):
        input_maps[n] = np.array([hp.alm2map(np.ascontiguousarray(input_alms[channel, :, c]), config.nside, lmax=config.lmax, pol=False) for c in range(input_alms.shape[-1])]).T

    output_maps = _get_diagnostic_maps(config, input_maps, compsep_run, np.ones(config.lmax+1), noise_debias=compsep_run["cov_noise_debias"])

    del compsep_run['good_channels']

    if "mask" in compsep_run:
        output_maps[compsep_run["mask"] == 0.] = 0.

    return output_maps

def _gilc_needlet(config: Configs, input_alms: np.ndarray, compsep_run: dict, **kwargs) -> np.ndarray:
    """
    Apply GILC in the needlet domain.

    Parameters
    ----------
        config: Configs
            Configuration object which includes general settings like nside and lmax. See `gilc` for details.
        input_alms: np.ndarray
            Input alms with shape (n_channels, n_alms, n_components).
        compsep_run: dict
            GILC parameters including needlet configuration. See `gilc` for details.
        **kwargs: dict
            Extra parameters for alm2map/map2alm.

    Returns
    ----------
        np.ndarray
            Output GNILC maps with shape (n_channels_out, npix, n_components).
    """
    b_ell = _get_needlet_windows_(compsep_run["needlet_config"], config.lmax)
    if compsep_run["b_squared"]:
        b_ell = b_ell**2

    if compsep_run['save_needlets']:
        compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        os.makedirs(compsep_run["path_out"], exist_ok=True)
        np.save(os.path.join(compsep_run["path_out"], "needlet_bands"), b_ell)
        
    output_alms = np.zeros((len(compsep_run["channels_out"]), input_alms.shape[1], input_alms.shape[-1]), dtype=complex)
    for j in range(b_ell.shape[0]):
        output_alms += _gilc_needlet_j(
            config, input_alms, compsep_run,
            b_ell[j], depro_cmb=compsep_run["depro_cmb"][j], m_bias=compsep_run["m_bias"][j], 
            noise_debias=compsep_run["cov_noise_debias"][j], **kwargs
        )

    output_maps = np.zeros((output_alms.shape[0], 12 * config.nside**2, output_alms.shape[-1]))
    for f, c in np.ndindex(output_alms.shape[0],output_alms.shape[-1]):
        output_maps[f,:,c] = hp.alm2map(np.ascontiguousarray(output_alms[f, :, c]), config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)
    
    if "mask" in compsep_run:
        output_maps[:,compsep_run["mask"] == 0.,:] = 0.

    return output_maps

def _fgd_diagnostic_needlet(config: Configs, input_alms: np.ndarray, compsep_run: dict) -> np.ndarray:
    """
    Compute diagnostic maps of foreground emission at different needlet scales.

    Parameters
    ----------
        config (Configs)
            Configuration object which includes general settings like nside and lmax. See `fgd_diagnostic` for details.
        input_alms: np.ndarray
            Input alms with shape (n_channels, n_alms).
        compsep_run: dict
            Dictionary with component separation parameters, including needlet configuration. See `fgd_diagnostic` for details.

    Returns
    -------
        np.ndarray
            Diagnostic needlet-space maps with shape (n_bands, npix).
    """
    b_ell = _get_needlet_windows_(compsep_run["needlet_config"], config.lmax)
    if compsep_run["b_squared"]:
        b_ell = b_ell**2

    if compsep_run['save_needlets']:
        compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        os.makedirs(compsep_run["path_out"], exist_ok=True)
        np.save(os.path.join(compsep_run["path_out"], "needlet_bands"), b_ell)
        
    output_maps = np.zeros((b_ell.shape[0], 12 * config.nside**2))
    for j in range(b_ell.shape[0]):
        output_maps[j] = _fgd_diagnostic_needlet_j(config, input_alms, compsep_run, b_ell[j], noise_debias=compsep_run["cov_noise_debias"][j])
    
    if "mask" in compsep_run:
        output_maps[:,compsep_run["mask"] == 0.] = 0.

    return output_maps

def _gilc_needlet_j(config: Configs, input_alms: np.ndarray, compsep_run: dict,
                    b_ell: np.ndarray, depro_cmb: Optional[float] = None, m_bias: Optional[int] = 0, 
                    noise_debias: Optional[float] = 0., **kwargs) -> np.ndarray:
    """
    Perform GNILC component separation in a specific needlet band.

    Parameters
    ----------
    
        config: Configs
            Configuration object which includes general settings like nside, lmax, units. See `gilc` for details.
        input_alms: np.ndarray
            Input alms for the scalar fields. Shape should be (n_channels, n_alms, n_components).
        compsep_run: dict
            Dictionary with component separation parameters. See `gilc` for details.
        b_ell: np.ndarray
            Needlet band window function for the current band. Shape should be (lmax+1,).
        depro_cmb: float, optional
            Deprojection factor for CMB in this needlet band. 
            Residual CMB in GNILC maps will be at the level of depro_CMB * CMB_input.
        m_bias: int, optional
            It will include m_bias more (if m_bias > 0) or less (if m_bias < 0) modes in the reconstructed GNILC maps.
        noise_debias: float, optional
            Noise debiasing factor. If set to a non-zero value, it will subtract a 'noise_debias' fraction of
            noise covariance from the input and nuisance covariance matrices.
        **kwargs: dict
            Additional keyword arguments for healpy function map2alm.

    Returns
    -------
        np.ndarray: 
            Output GNILC alms for the provided needlet band. Shape will be (n_channels_out, n_alms, n_components). 
    """

    if "mask" in compsep_run or not compsep_run["adapt_nside"]:
        nside_, lmax_ = config.nside, config.lmax
    else:
        nside_, lmax_ = _get_nside_lmax_from_b_ell(b_ell,config.nside,config.lmax)
        
    compsep_run["good_channels"] = _get_good_channels_nl(config, b_ell)

    input_maps_nl = np.zeros((compsep_run["good_channels"].shape[0], 12 * nside_**2, input_alms.shape[-1]))
    for n, channel in enumerate(compsep_run["good_channels"]):
        input_alms_j = _needlet_filtering(input_alms[channel], b_ell, lmax_)
        input_maps_nl[n] = np.array([
            hp.alm2map(np.ascontiguousarray(input_alms_j[:, c]), nside_, lmax=lmax_, pol=False)
            for c in range(input_alms.shape[-1])
        ]).T

    output_maps_nl = _gilc_maps(config, input_maps_nl, compsep_run, b_ell, depro_cmb=depro_cmb, m_bias=m_bias, noise_debias=noise_debias)
    del input_maps_nl

    output_alms_j = np.zeros((output_maps_nl.shape[0], hp.Alm.getsize(config.lmax), output_maps_nl.shape[-1]), dtype=complex)
    
    for n in range(output_maps_nl.shape[0]):
        output_alms_nl = np.array([
            hp.map2alm(output_maps_nl[n, :, c], lmax=lmax_, pol=False, **kwargs)
            for c in range(output_maps_nl.shape[-1])
        ]).T
        if compsep_run["b_squared"]:
            output_alms_j[n] = _needlet_filtering(output_alms_nl, np.ones(lmax_+1), config.lmax)
        else:
            output_alms_j[n] = _needlet_filtering(output_alms_nl, b_ell[:lmax_+1], config.lmax)

    del compsep_run['good_channels']

    return output_alms_j

def _fgd_diagnostic_needlet_j(config: Configs, input_alms: np.ndarray,
                               compsep_run: dict, b_ell: np.ndarray, noise_debias: Optional[float] = 0.
                            ) -> np.ndarray:
    """
    Compute diagnostic map of foreground complexity for a specific needlet band.

    Parameters
    ----------
        config: Configs
            Configuration object which includes general settings like nside and lmax. See `fgd_diagnostic` for details.
        input_alms: np.ndarray
            Input alms for the scalar fields. Shape should be (n_channels, n_alms).
        compsep_run: dict
            Dictionary with component separation parameters. See `fgd_diagnostic` for details.
        b_ell: np.ndarray
            Needlet band window function for the current band. Shape should be (lmax+1,).
        noise_debias: float, optional
            Noise debiasing factor. If set to a non-zero value, it will subtract a 'noise_debias' fraction of
            noise covariance from the input and nuisance covariance matrices.
    
    Returns
    -------
        np.ndarray
            Diagnostic map of foreground complexity for the provided needlet band.

    """
    if "mask" in compsep_run or not compsep_run["adapt_nside"]:
        nside_, lmax_ = config.nside, config.lmax
    else:
        nside_, lmax_ = _get_nside_lmax_from_b_ell(b_ell,config.nside,config.lmax)
        
    compsep_run["good_channels"] = _get_good_channels_nl(config, b_ell)
#    compsep_run["good_channels"] = np.arange(input_alms.shape[0])

    input_maps_nl = np.zeros((compsep_run["good_channels"].shape[0], 12 * nside_**2, input_alms.shape[-1]))
    for n, channel in enumerate(compsep_run["good_channels"]):
        input_alms_j = _needlet_filtering(input_alms[channel], b_ell, lmax_)
        input_maps_nl[n] = np.array([
            hp.alm2map(np.ascontiguousarray(input_alms_j[:, c]), nside_, lmax=lmax_, pol=False)
            for c in range(input_alms.shape[-1])
        ]).T

    output_maps_nl = _get_diagnostic_maps(config, input_maps_nl, compsep_run, b_ell, noise_debias=noise_debias)
    del input_maps_nl
    
    if hp.get_nside(output_maps_nl) < config.nside:
        output_maps_nl = hp.ud_grade(output_maps_nl, nside_out=config.nside)

    del compsep_run['good_channels']

    return output_maps_nl

def _gilc_maps(
    config: Configs,
    input_maps: np.ndarray,
    compsep_run: Dict[str, Any],
    b_ell: np.ndarray,
    depro_cmb: Optional[float] = None,
    m_bias: Union[int, float] = 0,
    noise_debias: Optional[float] = 0.,
) -> np.ndarray:
    """
    Apply GILC component separation to the provided scalar multifrequency maps.
    
    Parameters
    ----------
        config: Configs
            Configuration object which includes general settings like nside and lmax. See `gilc` for details.
        input_maps: np.ndarray
            Input maps for the scalar fields. Shape should be (n_channels, npix, n_components).
        compsep_run: Dict[str, Any]
            Dictionary with component separation parameters. See `gilc` for details.
        b_ell: np.ndarray
            Needlet band window function for the current band. Shape should be (lmax+1,).
        depro_cmb: float, optional
            Deprojection factor for CMB in this needlet band. 
            Residual CMB in GNILC maps will be at the level of depro_CMB * CMB_input.
        m_bias: int, optional
            It will include m_bias more (if m_bias > 0) or less (if m_bias < 0) modes in the reconstructed GNILC maps.
        noise_debias: float, optional
            Noise debiasing factor. If set to a non-zero value, it will subtract a 'noise_debias' fraction of
            noise covariance from the input and nuisance covariance matrices.
    
    Returns
    -------
        np.ndarray
            Output GNILC maps with shape (n_channels_out, npix, n_components).
    
    """
    # Compute covariance matrices for input and nuisance maps
    cov = (get_ilc_cov(input_maps[...,0], config.lmax, compsep_run, b_ell)).T
    if isinstance(compsep_run["nuis_idx"], int):
        cov_n = (get_ilc_cov(input_maps[...,compsep_run["nuis_idx"]], config.lmax, compsep_run, b_ell)).T
    elif isinstance(compsep_run["nuis_idx"], list):
        cov_n = (get_ilc_cov(input_maps[...,compsep_run["nuis_idx"][0]] + input_maps[...,compsep_run["nuis_idx"][1]], config.lmax, compsep_run, b_ell)).T

    if noise_debias != 0.:
        cov_noi = (get_ilc_cov(input_maps[...,compsep_run["noise_idx"]], config.lmax, compsep_run, b_ell)).T
        cov = cov - noise_debias * cov_noi
        cov_n = cov_n - noise_debias * cov_noi
        del cov_noi

    λ, U = Cn_C_Cn(cov,cov_n)
    λ[λ<1.]=1.

    W = _get_gilc_weights(config, U, λ, cov, cov_n, input_maps.shape, compsep_run, depro_cmb=depro_cmb, m_bias=m_bias) 

    del cov, cov_n, U, λ

    if compsep_run["ilc_bias"] == 0.:
        output_maps = np.einsum('li,ijk->ljk', W, input_maps)
    else:
        output_maps = np.einsum('jli,ijk->ljk', W, input_maps)

    outputs = []
    for channel in compsep_run["channels_out"]:
        if channel in compsep_run["good_channels"]:
            outputs.append(output_maps[compsep_run["good_channels"] == channel][0])
        else:
            outputs.append(np.zeros(output_maps.shape[1:]))
    del output_maps

    return np.array(outputs)

def _get_diagnostic_maps(
    config: Configs,
    input_maps: np.ndarray,
    compsep_run: dict,
    b_ell: np.ndarray,
    noise_debias: Optional[float] = 0.,
) -> np.ndarray:
    """
    Get diagnostic maps of foreground complexity for provided scalar field either in pixel or needlet domain.

    Parameters
    ----------
        config: Configs
            Configuration object which includes general settings like nside and lmax. See `fgd_diagnostic` for details.
        input_maps: np.ndarray
            Input maps for the scalar fields. Shape should be (n_channels, npix, n_comps).
        compsep_run: dict
            Dictionary with component separation parameters. See `fgd_diagnostic` for details.
        b_ell: np.ndarray
            Needlet band window function for the current band. Shape should be (lmax+1,).
        noise_debias: float, optional
            Noise debiasing factor. If set to a non-zero value, it will subtract a 'noise_debias' fraction of
            noise covariance from the input and nuisance covariance matrices.
    
    Returns
    -------
        np.ndarray
            Diagnostic map of foreground complexity for the provided scalar field.

    """

    cov = (get_ilc_cov(input_maps[...,0], config.lmax, compsep_run, b_ell)).T
    cov_n = (get_ilc_cov(input_maps[...,1], config.lmax, compsep_run, b_ell)).T
    
    if noise_debias != 0.:
        cov_noi = (get_ilc_cov(input_maps[...,2], config.lmax, compsep_run, b_ell)).T
        cov = cov - noise_debias * cov_noi
        cov_n = cov_n - noise_debias * cov_noi
        del cov_noi

    λ, U = Cn_C_Cn(cov,cov_n)
    λ[λ<1.]=1.
    del cov, cov_n, U

    m = _get_gilc_m(λ) 

    if isinstance(m, (int, float)):
        m = np.repeat(m, input_maps.shape[-2])
        if "mask" in compsep_run:
            m[compsep_run["mask"] == 0.] = 0.
        return m
    else:
        if "mask" in compsep_run:
            m_full = np.zeros(input_maps.shape[-2])
            m_full[compsep_run["mask"] > 0.] = m
            return m_full
        else:
            return m

def _get_gilc_weights(
    config: Configs,
    U: np.ndarray,
    λ: np.ndarray,
    cov: np.ndarray,
    cov_n: np.ndarray,
    input_shapes: tuple,
    compsep_run: dict,
    depro_cmb: Optional[float] = None,
    m_bias: Union[int, float] = 0
) -> np.ndarray:
    """
    Get GILC weights for the provided scalar field.

    Parameters
    ----------
        config: Configs
            Configuration object which includes general settings like nside and lmax. See `gilc` for details.
        U: np.ndarray
            Eigenvectors of the covariance matrix. Shape should be (n_channels, n_channels, n_pix).
        λ: np.ndarray
            Eigenvalues of the covariance matrix. Shape should be (n_channels, n_pix).
        cov: np.ndarray
            Covariance matrix of the input maps. Shape should be (n_channels, n_channels, n_pix).
        cov_n: np.ndarray
            Covariance matrix of the nuisance maps. Shape should be (n_channels, n_channels, n_pix).
        input_shapes: tuple
            Shapes of the input maps, typically (n_channels, npix, n_components).
        compsep_run: dict
            Dictionary with component separation parameters. See `gilc` for details.
        depro_cmb: float, optional
            Deprojection factor for CMB in this needlet band.
            Residual CMB in GNILC maps will be at the level of depro_CMB * CMB_input.
        m_bias: int, optional
            It will include m_bias more (if m_bias > 0) or less (if m_bias < 0) modes in the reconstructed GNILC maps.
    
    Returns
    -------
        np.ndarray
            GILC weights for the provided scalar field. 
            Shape will be (n_channels, n_channels) if cov is 2D, or (npix, n_channels, n_channels) if cov is 3D.
    """
    bandwidths = _get_bandwidths(config, compsep_run["good_channels"])
    A_cmb = _get_CMB_SED(np.array(config.instrument.frequency)[compsep_run["good_channels"]], 
                units=config.units, bandwidths=bandwidths)

    if cov.ndim == 2:
        inv_cov = lg.inv(cov)

        m = _get_gilc_m(λ)
        m += int(m_bias)  

        U_s = np.delete(U,np.where(λ < λ[m-1]),axis=1)

        F = scipy.linalg.sqrtm(cov_n) @ U_s
        
        if depro_cmb is None:
            W = F @ lg.inv(F.T @ inv_cov @ F) @ F.T @ inv_cov
        else:
#            F_e = np.column_stack([F, depro_cmb * np.ones(input_shapes[0])])
            F_e = np.column_stack([F, depro_cmb * A_cmb])
            F_A = np.column_stack([F, A_cmb])
            W = F_e @ lg.inv(F_A.T @ inv_cov @ F_A) @ F_A.T @ inv_cov

    elif cov.ndim == 3:
        m = _get_gilc_m(λ)
        m += int(m_bias)  

        covn_sqrt = lg.cholesky(cov_n)
#        covn_sqrt_inv = lg.inv(covn_sqrt)
        
        W_=np.zeros((cov.shape[0],input_shapes[0],input_shapes[0]))

        for m_ in np.unique(m):
            U_s = U[m==m_,:,:m_]
            cov_inv = lg.inv(cov[m==m_])
            F = np.einsum("kij,kjz->kiz", covn_sqrt[m==m_], U_s)

            if depro_cmb is None: 
                W_[m==m_] = np.einsum(
                    "kil,klj->kij", F,
                    np.einsum("kzl,klj->kzj",
                              lg.inv(np.einsum("kiz,kij,kjl->kzl", F, cov_inv, F)),
                              np.einsum("kiz,kij->kzj", F, cov_inv))
                )
            else:
                #e_cmb = depro_cmb * np.ones((F.shape[0],F.shape[1],1)) 
                #F_e = np.concatenate((F,e_cmb),axis=2)
                F_e = np.concatenate((F,np.tile(depro_cmb * A_cmb, (F.shape[0], 1))[:, :, np.newaxis]),axis=2)
                F_A = np.concatenate((F,np.tile(A_cmb, (F.shape[0], 1))[:, :, np.newaxis]),axis=2)
                W_[m==m_] = np.einsum(
                    "kil,klj->kij", F_e,
                    np.einsum("kzl,klj->kzj",
                              lg.inv(np.einsum("kiz,kij,kjl->kzl", F_A, cov_inv, F_A)),
                              np.einsum("kiz,kij->kzj", F_A, cov_inv))
                )
                del F_e, F_A #e_cmb, 

        del covn_sqrt, cov_inv, U_s, F, m

        if "mask" in compsep_run:
            W = np.zeros((input_shapes[1],W_.shape[1],W_.shape[2]))
            W[compsep_run["mask"] > 0.] = np.copy(W_)
        else:
            if W_.shape[0] != input_shapes[1]:
                W = np.zeros((input_shapes[1],W_.shape[1],W_.shape[2]))
                for i, k in np.ndindex(W_.shape[1],W_.shape[2]):
                    W[:,i,k]=hp.ud_grade(W_[:,i,k],hp.npix2nside(input_shapes[1]))
            else:
                W=np.copy(W_)
        del W_

    return W

def _get_gilc_m(λ: np.ndarray) -> Union[int, np.ndarray]:
    """
    Estimate the number of signal components using eigenvalue-based AIC.

    Parameters
    ----------
        λ: np.ndarray
            Eigenvalues of the whitened signal+nuisance covariance matrix.

    Returns
    -------
        int or np.ndarray
            Estimated number of components. If λ is 1D, returns a single integer.
            If λ is 2D, returns an array of integers of shape (npix).
    """

    if λ.ndim==1:
        A_m=np.zeros(λ.shape[0]+1)
        for i in range(λ.shape[0]):
            A_m[i]=2*i + np.sum(λ[i:]-np.log(λ[i:])-1.)
        A_m[λ.shape[0]]=2.*λ.shape[0]
        m = int(np.argmin(A_m))
    elif λ.ndim==2:
        A_m=np.zeros((λ.shape[0],λ.shape[1]+1))
        for i in range(λ.shape[1]):
            A_m[:,i]=2*i + np.sum(λ[:,i:]-np.log(λ[:,i:])-1.,axis=1)
        A_m[:,λ.shape[1]]=2.*λ.shape[1]
        m = np.argmin(A_m,axis=1)
    
    del A_m

    return m

def Cn_C_Cn(C: np.ndarray, C_n: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Whiten the signal covariance with nuisance and perform SVD.

    Parameters
    ----------
        C: np.ndarray
            Signal covariance matrix. Shape should be (n_channels, n_channels) or (npix, n_channels, n_channels).
        C_n: np.ndarray
            Nuisance covariance matrix. Shape should be (n_channels, n_channels) or (npix, n_channels, n_channels).

    Returns
    -------
        tuple[np.ndarray, np.ndarray]
            Tuple of eigenvalues and eigenvectors.
    """
    if (C.ndim == 2) and (C_n.ndim == 2):
        Cn_sqrt = lg.inv(scipy.linalg.sqrtm(C_n))
        M = lg.multi_dot([Cn_sqrt,C,Cn_sqrt])
    elif (C.ndim == 3) and (C_n.ndim == 3):
        Cn_sqrt = lg.inv(lg.cholesky(C_n))
        M = np.einsum("kij,kjz->kiz", Cn_sqrt, np.einsum("kij,kzj->kiz", C, Cn_sqrt))
    U, λ, _ = lg.svd(M)
    return λ, U

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]
                    

   
    